"""A module for loading :class:`WorkFlow` templates."""

import os
from types import MappingProxyType
from typing import Mapping, MutableMapping, Tuple, Dict, Any, TYPE_CHECKING

import yaml
import numpy as np

from .key_map import (
    OPT,
    MOL,
    HDF5_INDEX,
    JOB_SETTINGS_QD_OPT,
    JOB_SETTINGS_CRS,
    JOB_SETTINGS_BDE,
    JOB_SETTINGS_ASA,
    JOB_SETTINGS_CDFT,
    ASA_INT,
    ASA_STRAIN,
    ASA_E,
    SETTINGS1,
    SETTINGS2,
    SETTINGS_SOLV1,
    SETTINGS_SOLV2,
    SETTINGS_ASA,
    SETTINGS_BDE1,
    SETTINGS_BDE2,
    SETTINGS_CDFT,
    V_BULK,
    CDFT_MU,
    CDFT_CHI,
    CDFT_ETA,
    CDFT_S,
    CDFT_GAMMA,
    CDFT_OMEGA,
    CDFT_NUCLEOFUGE,
    CDFT_ELECTROFUGE,
    CDFT_W_MINUS,
    CDFT_W_PLUS,
    CDFT_ELECTROPHILICITY,
    CDFT_DELTAF_MINUS,
    CDFT_DELTAF_PLUS,
    CDFT_MU_MINUS,
    CDFT_MU_PLUS
)

if TYPE_CHECKING:
    from nanoutils import TypedDict

    class _TemplateMapping(TypedDict):
        description: str
        mol_type: str
        template: Mapping[str, Tuple[str, ...]]
        import_columns: Mapping[Tuple[str, str], Any]
        export_columns: Tuple[Tuple[str, str], ...]

else:
    _TemplateMapping = 'CAT.workflows.workflow_dicts._TemplateMapping'

__all__ = ['WORKFLOW_TEMPLATE']


def _load_templates() -> Dict[str, Any]:
    """Load the templates from ``CAT/data/workflow_dicts/workflow_yaml.yaml``."""
    path = os.path.join(os.path.dirname(__file__), 'workflow_yaml.yaml')
    with open(path, 'r') as f:
        return yaml.load(f, Loader=yaml.FullLoader)


def _recursive_mapping_proxy(dct: MutableMapping) -> MappingProxyType:
    """Recursivelly convert **dct**, and all nested mutable mappings, into :class:`MappingProxyType<types.MappingProxyType>` instances."""  # noqa
    for k, v in dct.items():
        if isinstance(v, dict):
            dct[k] = _recursive_mapping_proxy(v)
    return MappingProxyType(dct)


def finalize_templates():
    """Update the templates using :data:`._TEMPLATE_UPDATE`."""
    base_dict = {
        'asa': {'import_columns': {ASA_INT: np.nan, ASA_STRAIN: np.nan, ASA_E: np.nan},
                'export_columns': (JOB_SETTINGS_ASA, SETTINGS_ASA, ASA_INT, ASA_STRAIN, ASA_E)},
        'ligand_opt': {'import_columns': {HDF5_INDEX: -1, OPT: False},
                       'export_columns': (MOL, HDF5_INDEX, OPT, SETTINGS1, SETTINGS2)},
        'qd_attach': {'import_columns': {HDF5_INDEX: -1, OPT: False},
                      'export_columns': (MOL, HDF5_INDEX)},
        'qd_opt': {'import_columns': {HDF5_INDEX: -1, OPT: False},
                   'export_columns': (MOL, HDF5_INDEX, OPT, JOB_SETTINGS_QD_OPT,
                                      SETTINGS1, SETTINGS2)},
        'crs': {'import_columns': {},
                'export_columns': (JOB_SETTINGS_CRS, SETTINGS_SOLV1, SETTINGS_SOLV2)},
        'bde': {'import_columns': {},
                'export_columns': (JOB_SETTINGS_BDE, SETTINGS_BDE1, SETTINGS_BDE2)},
        'forcefield': {'import_columns': {},
                       'export_columns': ()},
        'bulkiness': {'import_columns': {V_BULK: np.nan},
                      'export_columns': (V_BULK,)},
        'multi_ligand': {'import_columns': {HDF5_INDEX: -1, OPT: False},
                         'export_columns': (HDF5_INDEX, OPT)},
        'cdft': {'import_columns': {CDFT_MU: np.nan, CDFT_CHI: np.nan, CDFT_ETA: np.nan,
                                    CDFT_S: np.nan, CDFT_GAMMA: np.nan, CDFT_OMEGA: np.nan,
                                    CDFT_NUCLEOFUGE: np.nan, CDFT_ELECTROFUGE: np.nan,
                                    CDFT_W_MINUS: np.nan, CDFT_W_PLUS: np.nan,
                                    CDFT_ELECTROPHILICITY: np.nan, CDFT_DELTAF_MINUS: np.nan,
                                    CDFT_DELTAF_PLUS: np.nan, CDFT_MU_MINUS: np.nan,
                                    CDFT_MU_PLUS: np.nan},
                 'export_columns': (JOB_SETTINGS_CDFT, SETTINGS_CDFT, CDFT_MU, CDFT_CHI, CDFT_ETA,
                                    CDFT_S, CDFT_GAMMA, CDFT_OMEGA, CDFT_NUCLEOFUGE,
                                    CDFT_ELECTROFUGE, CDFT_W_MINUS, CDFT_W_PLUS,
                                    CDFT_ELECTROPHILICITY, CDFT_DELTAF_MINUS, CDFT_DELTAF_PLUS,
                                    CDFT_MU_MINUS, CDFT_MU_PLUS)},
    }

    templates = _load_templates()
    for k, v1 in templates.items():
        v2 = base_dict[k]
        v1.update(v2)

        # Convert all lists stored under the 'template' key into tuples
        for key, list_ in v1['template'].items():
            v1['template'][key] = tuple(list_)
    return _recursive_mapping_proxy(templates)


#: An immutable mapping with additional values for ``CAT/data/workflow_dicts/workflow_yaml.yaml``.
#: Contains values which are generally not as easily represented in the .yaml format.
WORKFLOW_TEMPLATE: Mapping[str, _TemplateMapping] = finalize_templates()




"""A module for creating new functions from a strings.

Index
-----
.. currentmodule:: CAT.data_handling.str_to_func
.. autosummary::
    str_to_func
    compile_func

API
---
.. autofunction:: str_to_func
.. autofunction:: compile_func

"""
from types import FunctionType, CodeType
from typing import Union, Callable

import numpy as np

from .warn_map import WARN_MAP

__all__ = ['str_to_func']


def str_to_func(func_str: Union[Callable[[np.ndarray], np.ndarray], str],
                shape: str = 'raise', dtype: str = 'raise') -> Callable[[np.ndarray], np.ndarray]:
    """Convert a string-representation of a function into an actual function and validate its output."""  # noqa
    shape_func = WARN_MAP[shape]
    dtype_func = WARN_MAP[dtype]

    # Create the new function
    if callable(func_str):  # **func_str** is already a function; no further compilation required
        func = func_str
    else:
        func = compile_func(func_str)

    # Validate and return the function
    ar1 = np.array([[0.0, np.nan], [np.inf, 5.0]])
    ar2 = func(ar1)
    if ar1.shape != ar2.shape:
        shape_func(ValueError(f"{func_str} -> y: input and output arrays have "
                              f"non-identical shapes: {ar1.shape} and {ar2.shape}"))
    if ar1.dtype != ar2.dtype:
        dtype_func(ValueError(f"{func_str} -> y: input and output arrays have non-identical "
                              f"data types: '{ar1.dtype.name}' and '{ar2.dtype.name}'"))
    return func


def compile_func(func_str: str) -> FunctionType:
    """Compile **func_str** into a :class:`code<types.CodeType>` instance."""
    global_dict = {'np': np}
    x = np.array([[0.0, np.nan], [np.inf, 5.0]])

    func_str = func_str.replace('numpy', 'np')

    # Load all module specified in **func_str**
    while True:
        try:
            x.copy()
            exec(func_str)
        except NameError as ex:
            module = str(ex).split("'")[1]
            exec(f'import {module}')
            global_dict[module] = eval(module)
        else:
            break

    # Exchange all ';' characters for '\n'
    str_list = func_str.split(';')
    func_str2 = ''.join(f'    {i.strip().rstrip()}\n' for i in str_list[:-1])
    func_str2 += f'    return {str_list[-1].strip().rstrip()}'

    # Compile the code of the to-be returned function
    code_compile = compile(f'def weight(x):\n{func_str2}', '<string>', 'exec')
    for code in code_compile.co_consts:
        if isinstance(code, CodeType):
            break
    else:
        raise ValueError("Failed to find 'code' instance in 'code_compile.co_consts'")

    # Construct and return the new function
    func = FunctionType(code, globals=global_dict, name='weight')
    func.__doc__ = f"Dynamically generated function; returns ``{func_str}``."
    func.__annotations__ = {'x': np.ndarray, 'return': np.ndarray}
    return func



"""A module to hold the :class:`GenJobManager` class.

Index
-----
.. currentmodule:: CAT.gen_job_manager
.. autosummary::
    GenJobManager

API
---
.. autoclass:: GenJobManager
   :members:
   :private-members:
   :special-members:

"""

import os
import errno
from typing import (Optional, Callable)
from os.path import (join, isfile, abspath, isdir, exists, normpath)

import dill as pickle

from scm.plams import (JobManager, FileError, Settings, PlamsError)
from scm.plams.core.basejob import Job

from .logger import logger

__all__ = ['GenJobManager']


class GenJobManager(JobManager):
    """A modified version of the PLAMS :class:`JobManager` class.

    In order to minimize its memory footprint, :meth:`GenJobManager.load_job` will now populate
    the values of :attr:`GenJobManager.hashes` with a callable the creates and returns a
    :class:`Job` instance (rather than permantly sotring the :class:`Job` instance).

    Attributes
    ----------
    foldername : str
        The name of the working folder (see :attr:`GenJobManager.workdir`).

    workdir : str
        The absolute path to the working folder.

    logfile : str
        The absolute path+filename of the logfile.

    input : str
        The absolute path to the copy of the input file in the working folder.

    settings : |plams.Settings|_
        A :class:`Settings` instance with settings for this job manager.

    jobs : |list|_ [|plams.Jobs|_]
        A list of all jobs managed with this instance (in order of run calls).

    names : |dict|_ [|str|_, |int|_]
        A dictionary with names of jobs.
        For each name an integer value is stored indicating
        how many jobs with that basename have already been run.

    hashes : |dict|_ [|str|_, |Callable|_]
        A dictionary working as a hash-table for jobs.
        Values created by :meth:`GenJobManager.load_job` are stored as callables which
        in turn create :class:`Job` instances.

    """

    def __init__(self, settings: Settings,
                 path: Optional[str] = None,
                 folder: Optional[str] = None,
                 hashing: str = 'input') -> None:
        """Initialize the :class:`GenJobManager` instance."""
        self.settings = settings
        self.jobs = []
        self.names = {}
        self.hashes = {}

        if path is None:
            self.path = os.getcwd()
        elif isdir(path):
            self.path = abspath(path)
        else:
            raise PlamsError(f'Invalid path: {path}')

        basename = normpath(folder) if folder else 'plams_workdir'
        self.foldername = basename
        self.workdir = join(self.path, self.foldername)
        self.logfile = join(self.workdir, 'logfile')
        self.input = join(self.workdir, hashing)
        if not exists(self.workdir):
            os.mkdir(self.workdir)

    @staticmethod
    def _unpickle(filename: str) -> Job:
        """Attempt to unpickle and return a :class:`Job` containing file."""
        with open(filename, 'rb') as f:
            try:
                return pickle.load(f)
            except Exception as ex:
                raise FileError(f'Failed to unpickle {filename!r}') from ex

    def _get_job(self, filename: str) -> Callable[[], Job]:
        """Return a callable which converts **filename** into a :class:`Job` instance."""
        def unpickle_job() -> Job:
            _filename = filename.replace('.hash', '.dill')
            ret = GenJobManager._unpickle(_filename)
            ret.jobmanager = self
            return ret
        return unpickle_job

    def load_job(self, filename: str) -> None:
        """Load a previously saved job from **filename**, populating :attr:`GenJobManager.hashes`.

        The hash of the loaded job is stored as key in :attr:`GenJobManager.hashes`,
        the matching value being a callable that loads the actual the :class:`Job` instance.

        Parameters
        ----------
        filename : str
            A path to a .hash file in some job folder.
            A :class:`Job` instance stored there is loaded and returned.
            All attributes of this instance removed before pickling are restored.

        """
        # Raise an error if **filename** cannot be found
        if not isfile(filename):
            raise FileError(f'File {filename} not present')

        # Read the has from filename
        with open(filename, 'r') as f:
            h = f.read().rstrip('\n')
        self.hashes[h] = self._get_job(filename)

    def remove_job(self, job: Job) -> None:
        """Remove **job** from the job manager; forget its hash."""
        if job in self.jobs:
            self.jobs.remove(job)
            job.jobmanager = None
        h = job.hash()
        if h in self.hashes:
            del self.hashes[h]

    def _check_hash(self, job: Job) -> Optional[Job]:
        """Calculate and check the hash of **job**.

        If the hash is not ``None``, previous jobs are searched for the same hash.
        Matches are returned if applicable; returns ``None`` otherwise.

        Parameter
        ---------
        job : |plams.Job|_
            A :class:`Job` instance.

        Returns
        -------
        |plams.Job|_
            Optional: A matching :class:`Job` instance stored in :attr:`GenJobManager.hashes`.
            If no matching hash is found return ``None``.

        """
        h = job.hash()
        setattr(job, '_hash', h)
        if h is None:  # No hash available for **job**, move along
            return None

        if h in self.hashes:
            try:
                ret = self.hashes[h]()
                if not os.path.isdir(ret.path):
                    raise FileNotFoundError(errno.ENOENT, "No such file or directory", ret.path)
            except Exception as ex:  # In case the job unpickling fails
                logger.warning(f"Failed to unpickle {job.name!r}", exc_info=ex)
            else:
                return ret

        filename = join(job.path, job.name)
        func = self._get_job(filename + '.dill')
        self.hashes[h] = func  # Set a callable that returns a Job instance
        return None



"""A module designed for finding ligand functional groups.

Index
-----
.. currentmodule:: CAT.attachment.ligand_anchoring
.. autosummary::
    init_ligand_anchoring
    get_functional_groups
    _smiles_to_rdmol
    find_substructure
    substructure_split
    _get_df

API
---
.. autofunction:: init_ligand_anchoring
.. autofunction:: get_functional_groups
.. autofunction:: find_substructure
.. autofunction:: _smiles_to_rdmol
.. autofunction:: substructure_split
.. autofunction:: _get_df

"""

from itertools import chain, repeat
from collections import defaultdict
from typing import Sequence, List, Tuple, Iterable, Callable, Iterator

import pandas as pd

from scm.plams import Molecule, Settings, MoleculeError
import scm.plams.interfaces.molecule.rdkit as molkit

from rdkit import Chem

from ..logger import logger
from ..utils import get_template, AnchorTup, KindEnum, get_formula, FormatEnum
from ..mol_utils import separate_mod   # noqa: F401
from ..workflows import MOL, FORMULA, HDF5_INDEX, OPT
from ..settings_dataframe import SettingsDataFrame
from ..data_handling.validate_mol import santize_smiles

__all__ = ['init_ligand_anchoring']


def init_ligand_anchoring(ligand_df: SettingsDataFrame) -> SettingsDataFrame:
    """Initialize the ligand functional group searcher.

    Parameters
    ----------
    ligand_df : |CAT.SettingsDataFrame|_
        A dataframe of valid ligands.

    Returns
    -------
    |CAT.SettingsDataFrame|_
        A dataframe of ligands with functional groups that can serve as valid anchor points.

    """
    # Unpack arguments
    settings = ligand_df.settings.optional
    functional_groups = settings.ligand.anchor

    # Find all functional groups; return a copy of each mol for each functional group
    mol_list = []
    for lig in ligand_df[MOL]:
        # Functional group search
        dummies = lig.properties.dummies
        if not dummies:
            mol_list += find_substructure(lig, functional_groups)
            continue
        else:
            dummies = tuple(i - 1 for i in dummies)

        # Manual specification of a functional group
        if len(dummies) == 1:  # optional.ligand.split = False
            remove = None
        elif len(dummies) == 2:  # optional.ligand.split = True
            remove = (1,)
        else:
            raise NotImplementedError
        anchor_tup = AnchorTup(None, group_idx=(0,), remove=remove)
        mol_list += [substructure_split(lig, dummies, anchor_tup)]

    # Convert the results into a dataframe
    return _get_df(mol_list, ligand_df.settings)


def _get_df(
    mol_list: Sequence[Molecule],
    settings: Settings,
) -> SettingsDataFrame:
    """Create and return a new ligand dataframe.

    Parameters
    ----------
    mol_list : |list|_ [|plams.Molecule|_]
        A list of PLAMS molecules.
    settings : |Settings|_
        A Settings instance containing all CAT parameters.

    Returns
    -------
    |CAT.SettingsDataFrame|_
        A dataframe of ligands with functional groups that can serve as valid anchor points.

    """
    # Create the dataframe index and columns
    idx_tuples = [(mol.properties.smiles, mol.properties.anchor) for mol in mol_list]
    idx = pd.MultiIndex.from_tuples(idx_tuples, names=['smiles', 'anchor'])
    columns_tuples = [MOL, FORMULA, HDF5_INDEX, OPT]
    columns = pd.MultiIndex.from_tuples(columns_tuples, names=['index', 'sub index'])

    # Create, fill and return the dataframe
    df = SettingsDataFrame(-1, index=idx, columns=columns, settings=settings)
    df[MOL] = mol_list
    df[FORMULA] = [get_formula(lig) for lig in df[MOL]]
    df[OPT] = False
    return df[~df.index.duplicated(keep='first')]  # Remove duplicate indices


def get_functional_groups(
    functional_groups: "None | Iterable[str]" = None,
    split: bool = True,
) -> Tuple[Chem.Mol, ...]:
    """Construct a list of RDKit molecules representing functional groups.

    Parameters
    ----------
    functional_groups : |list|_ [|str|_]
        Optional: A list of smiles strings representing functional groups.
        Will default to templates provided by CAT if ``None``.
    split : bool
        If templates should be pulled from the ``['split']`` or ``['no_split']`` block.
        Only relevant if **functional_groups** is ``None``.

    Returns
    -------
    |tuple|_ [|Chem.Mol|_]
        A list of RDKit molecules constructed from either **functional_group** or
        the default smiles templates in CAT.

    """
    # The user has, explicitly, provided functional groups
    if functional_groups:
        return tuple(_smiles_to_rdmol(smiles) for smiles in functional_groups)

    # Read functional groups from the default CAT SMILES templates
    if split:
        func_groups = get_template('smiles.yaml').split
    else:
        func_groups = get_template('smiles.yaml').no_split

    return tuple(_smiles_to_rdmol(smiles) for smiles in func_groups)


_smiles_to_rdmol = FormatEnum.SMILES.value


def find_substructure(
    ligand: Molecule,
    func_groups: Iterable[AnchorTup],
    split: bool = True,
    condition: "None | Callable[[int], bool]" = None,
) -> List[Molecule]:
    """Identify interesting functional groups within the ligand.

    Parameters
    ----------
    ligand : |plams.Molecule|_
        The ligand molecule.
    func_groups : |tuple|_ [|Chem.Mol|_]
        A collection of RDKit molecules representing functional groups.

    Returns
    -------
    |list|_ [|plams.Molecule|_]
        A list of ligands.
        A single copy of **ligand** is created for each identified functional group,
        removing parts of the functional group if required (see **split**).
        An empty list is returned if no valid functional groups are found.

    """
    rdmol = molkit.to_rdmol(ligand)

    # Searches for functional groups (defined by functional_group_list) within the ligand
    matches: Iterator[Tuple[AnchorTup, Tuple[int, ...]]] = chain.from_iterable(
        zip(repeat(anchor_tup), rdmol.GetSubstructMatches(anchor_tup.mol, useChirality=True))
        for anchor_tup in func_groups
    )

    # Remove all duplicate matches, each heteroatom (match[0]) should have <= 1 entry
    ligand_idx_dict = defaultdict(list)
    ref_dict = defaultdict(set)
    for anchor_tup, idx_tup in matches:
        ref_set = ref_dict[anchor_tup]
        anchor_idx_tup = tuple(idx_tup[i] for i in anchor_tup.group_idx)
        if anchor_idx_tup in ref_set:
            continue  # Skip duplicates

        ligand_idx_dict[anchor_tup].append(idx_tup)
        ref_set.add(anchor_idx_tup)

    if condition is not None:
        if not condition(sum((len(i) for i in ligand_idx_dict.values()), 0)):
            err = (f"Failed to satisfy the passed condition ({condition!r}) for "
                   f"ligand: {ligand.properties.name!r}")
            logger.error(err)
            return []

    ret = []
    idx_dict_items = chain.from_iterable(zip(repeat(k), v) for k, v in ligand_idx_dict.items())
    for i, (anchor_tup, idx_tup) in enumerate(idx_dict_items):
        try:
            value = substructure_split(ligand, idx_tup, anchor_tup)
        except Exception as ex:
            logger.warning(
                f"Failed to parse {ligand.properties.name!r} anchoring group {i}", exc_info=ex
            )
        else:
            ret.append(value)

    if not ret:
        err = (f"No functional groups were found (optional.ligand.split = {split!r}) for "
               f"ligand: {ligand.properties.name!r}")
        logger.error(err)
    return ret


def substructure_split(
    ligand: Molecule,
    idx_tup: Tuple[int, ...],
    anchor_tup: AnchorTup,
) -> Molecule:
    """Delete the hydrogen or mono-/polyatomic counterion attached to the functional group.

    Sets the charge of the remaining heteroatom to -1 if ``split=True``.

    Parameters
    ----------
    ligand: plams.Molecule
        The ligand molecule.
    idx_tup : tuple[int, ...]
        A tuple with 2 atomic indices associated with a functional group.
    anchor_tup : AnchorTup
        Named tuple.

    Returns
    -------
    plams.Molecule
        A copy of **ligand**, with part of its functional group removed (see **split**).

    """
    lig = ligand.copy()
    anchors = [lig[1 + idx_tup[i]] for i in anchor_tup.group_idx]
    anchor = anchors[0]

    if anchor_tup.remove is not None:
        remove_iter = sorted([1 + idx_tup[i] for i in anchor_tup.remove], reverse=True)
        for i in remove_iter:
            lig.delete_atom(lig[i])

        mol_list: Tuple[Molecule, ...] = lig.separate_mod()
        for mol in mol_list:
            if anchor not in mol:
                continue
            lig = mol
            break
        else:
            raise MoleculeError("Failed to identify an anchor-containing fragment")

        # Check if the ligand heteroatom has a charge assigned, assigns a charge if not
        if not anchor.properties.charge:
            anchor.properties.charge = -1

    anchor_tup = anchor_tup._replace(anchor_idx=tuple(lig.atoms.index(at) for at in anchors))

    # Update ligand properties
    lig.properties.dummies = anchor
    if anchor_tup.kind == KindEnum.FIRST:
        i = 1 + anchor_tup.anchor_idx[0]
        lig.properties.anchor = f"{lig[i].symbol}{i}"
    else:
        lig.properties.anchor = "".join(
            f"{lig[1 + i].symbol}{1 + i}" for i in anchor_tup.anchor_idx
        )
    lig.properties.anchor_tup = anchor_tup
    lig.properties.charge = sum(atom.properties.get('charge', 0) for atom in lig)

    # Update the ligand smiles string
    rdmol = molkit.to_rdmol(lig)
    smiles = Chem.MolToSmiles(rdmol)  # NOTE: can return `None` under rare circumstances
    lig.properties.smiles = Chem.CanonSmiles(smiles)
    lig.properties.name = santize_smiles(lig.properties.smiles) + '@' + lig.properties.anchor
    lig.properties.path = ligand.properties.path
    return lig



"""A context manager for temporary splitting (and reassembling) molecules.

Index
-----
.. currentmodule:: CAT.attachment.mol_split_cm
.. autosummary::
    SplitMol

API
---
.. autoclass:: SplitMol
    :members:
    :private-members:
    :special-members:

"""

import copy
import reprlib
from typing import Iterable, Union, Dict, Tuple, NoReturn, Any, Type
from contextlib import AbstractContextManager

from scm.plams import Molecule, Atom, PT, Bond, MoleculeError, PTError, rotation_matrix

from ..mol_utils import separate_mod  # noqa: F401

__all__ = ['SplitMol']


class SplitMol(AbstractContextManager):
    """A context manager for temporary splitting a single molecule into multiple components.

    The context manager splits the provided molecule into multiple components,
    capping all broken bonds in the process.
    The exact amount of fragments depends on the number of specified bonds.

    These moleculair fragments are returned upon opening the context manager and merged back
    into the initial molecule once the context manager is closed. While opened, the initial molecule
    is cleared of all atoms and bonds, while the same hapens to the moleculair fragments upon
    closing.

    Examples
    --------
    .. code:: python

        >>> from scm.plams import Molecule, Bond, from_smiles

        >>> mol: Molecule = from_smiles('CC')  # Ethane
        >>> bond: Bond = mol[1, 2]

        # A backup of all bonds and atoms
        >>> bonds_backup = mol.bonds.copy()
        >>> atoms_backup = mol.atoms.copy()

        # The context manager is opened; the bond is removed and the molecule is fragmented
        >>> with SplitMol(mol, bond) as fragment_tuple:
        ...     for fragment in fragment_tuple:
        ...         fancy_operation(fragment)  # doctest: +SKIP
        ...
        ...     print(
        ...         mol.bonds == bonds_backup,
        ...         mol.atoms == atoms_backup,
        ...         bond in mol.bonds
        ...     )
        False False False

        # The context manager is closed; all atoms and bonds have been restored
        >>> print(
        ...     mol.bonds == bonds_backup,
        ...     mol.atoms == atoms_backup,
        ...     bond in mol.bonds
        ... )
        True True True

    Parameters
    ----------
    mol : |plams.Molecule|
        A PLAMS molecule.
        See :attr:`SplitMol.mol`.

    bond_list : |plams.Bond| or |Iterable| [|plams.Bond|]
        An iterable consisting of PLAMS bonds.
        All bonds must be part of **mol**.
        See :attr:`SplitMol.bonds`.

    cap_type : :class:`str`, :class:`int` or |plams.Atom|
        An atomic number or symbol of the atom type used for capping the to-be split molecule.
        See :attr:`SplitMol.cap_type`.

    Attributes
    ----------
    mol : |plams.Molecule|
        A PLAMS molecule.

    bonds : :class:`set` [|plams.Bond|]
        A set of PLAMS bonds.

    cap_type : :class:`str`
        An atomic symbol of the atom type used for capping the to-be split molecule.

    _at_pairs : :class:`list` [:class:`dict` [|plams.Atom|, |plams.Atom|]], optional
        A list of dictionaries.
        Each dictionary contains two atoms as keys (see :attr:`SplitMol.bond_list`) and
        their respective capping atom as values.
        Used for reassembling :attr:`SplitMol.mol` once the context manager is closed.
        Set internally by :meth:`SplitMol.__enter__`.

    _vars_backup : :class:`dict` [:class:`str`, |Any|], optional
        A backup of all instance variables of :attr:`SplitMol.mol`.
        Set internally by :meth:`SplitMol.__enter__`.

    _tmp_mol_list : :class:`tuple` [|plams.Molecule|], optional
        A list of PLAMS molecules obtained by splitting :attr:`SplitMol.mol`.
        Set internally by :meth:`SplitMol.__enter__`.

    Raises
    ------
    |MoleculeError|
        Raised when one attempts to access or manipulate the instance variables of
        :attr:`SplitMol.mol` when the context manager is opened.

    """

    def __init__(self, mol: Molecule, bond_list: Union[Bond, Iterable[Bond]],
                 cap_type: Union[str, int, Atom] = 'H') -> None:
        """Initialize a :class:`SplitMol` instance."""
        self.mol = mol
        self.bonds = bond_list
        self.cap_type = cap_type

        # Private variables which are modified in __enter__ and __exit__
        self._at_pairs = None
        self._vars_backup = None
        self._tmp_mol_list = None

    """####################################### Properties #######################################"""

    @property
    def bonds(self) -> Dict[Bond, None]:
        """Getter: Return :attr:`SplitMol.bonds`. Setter: Assign the provided value as a dictionary with bonds as keys and their matching index as value."""  # noqa
        return self._bonds

    @bonds.setter
    def bonds(self, value: Union[Bond, Iterable[Bond]]) -> None:
        if isinstance(value, Bond):
            self._bonds = {value: None}
        else:
            self._bonds = {i: None for i in value}

    @property
    def cap_type(self) -> str:
        """Getter: Return :attr:`SplitMol.cap_type`. Setter: Assign the provided value after parsing the atom type (:class:`str`)."""  # noqa
        return self._cap_type

    @cap_type.setter
    def cap_type(self, value: Union[str, int, Atom]) -> None:
        if isinstance(value, int):
            self._cap_type = PT.get_symbol(value)

        elif isinstance(value, Atom):
            self._cap_type = value.symbol

        elif isinstance(value, str):
            if value.capitalize() not in PT.symtonum:
                raise PTError(f"Trying to convert an invalid atomic symbol: {reprlib.repr(value)}")
            self._cap_type = value

        else:
            raise TypeError(f"Invalid atom type: {repr(type(value))}")

    """Methods for manipulating the molecule."""

    def cap_fragments(self, bond: Bond) -> Dict[Atom, Atom]:
        """Delete a bond from :attr:`SplitMol.mol` and cap the resulting fragments with :attr:`SplitMol.cap_type`.

        Parameters
        ----------
        bond : |plams.Bond|
            A PLAMS bond.

        Returns
        -------
        :class:`dict` [|plams.Atom|, |plams.Atom|]
            A dictionary with the old atoms in **bond** as keys and their new capping atoms
            as values.

        """  # noqa
        mol = self.mol
        symbol = self.cap_type

        # Construct the capping atoms
        atom1, atom2 = bond.atom1, bond.atom2
        atom1_cap = Atom(symbol=symbol, coords=atom2.coords)
        atom2_cap = Atom(symbol=symbol, coords=atom1.coords)
        mol.add_atom(atom1_cap, adjacent=[atom1])
        mol.add_atom(atom2_cap, adjacent=[atom2])

        # Resize the capping atom bond
        length1 = atom1.radius + atom1_cap.radius
        length2 = atom2.radius + atom2_cap.radius
        mol.bonds[-2].resize(atom1_cap, length1)
        mol.bonds[-1].resize(atom2_cap, length2)

        # Delete the old bond and return a dictionary containg marking all new bonds
        return {atom1: atom1_cap, atom2: atom2_cap}

    def reassemble(self) -> None:
        """Reassemble :attr:`SplitMol.mol` from its consitituent components in :attr:`SplitMol._tmp_mol_list`.

        All capping atoms (and their respective bonds) specified in :attr:`SplitMol._at_pairs` are
        removed and the previously broken bonds (stored in :attr:`SplitMol.bonds`) are restored.

        """  # noqa
        mol = self.mol
        mark = self._at_pairs
        bonds = self.bonds

        _iterator = enumerate(zip(mark, bonds), start=(1 - len(bonds)))
        for i, (atom_dict, bond) in _iterator:
            # Extract atoms
            iterator = iter(atom_dict.items())
            atom1, atom1_cap = next(iterator)
            atom2, atom2_cap = next(iterator)

            atom1_cap.bonds[0].resize(atom1_cap, atom1.radius + atom2.radius)
            atom2_cap.bonds[0].resize(atom2_cap, atom1.radius + atom2.radius)

            # Allign the molecules by rotation
            vec1 = atom2.vector_to(atom2_cap)
            vec2 = atom1_cap.vector_to(atom1)
            rotmat = rotation_matrix(vec1, vec2)
            atom2.mol.rotate(rotmat)

            # Allign the molecules by translation
            vec_trans = atom2.vector_to(atom1_cap)
            atom2.mol.translate(vec_trans)

            # Replace the capping atom bonds with the previously broken bond
            atom1.bonds[-1] = bond
            atom2.bonds[-1] = bond

            # Don't bother transfering ownsership for the last fragment
            if i:
                _tmp_mol = atom1.mol
                _tmp_mol.atoms += atom2.mol.atoms
                _tmp_mol.bonds += atom2.mol.bonds
                for at in atom2.mol.atoms:
                    at.mol = _tmp_mol
                for bond in atom2.mol.bonds:
                    bond.mol = _tmp_mol

        # Ensure all atoms and bonds belong to mol
        for at in mol.atoms:
            at.mol = mol
        for bond in mol.bonds:
            bond.mol = mol

        # Resize bonds
        for atom_dict, bond in zip(mark, bonds):
            atom1, atom2 = atom_dict
            length = atom1.radius + atom2.radius
            bond.resize(atom1, length)

    def lock_mol(self) -> None:
        """Lock :attr:`SplitMol.mol`, preventing any access to the instance."""
        mol = self.mol
        mol.atoms = mol.bonds = FrozenList()

    def unlock_mol(self) -> None:
        """Unlock :attr:`SplitMol.mol`, restoring access to the instance."""
        self.mol.__dict__ = self._vars_backup

    @staticmethod
    def reset_vars(obj: Any) -> None:
        """Replace all instance variables of **obj** with empty instances of their respective class.

        A value will be substituted for ``None`` if a :exc:`TypeError` is encountered during
        instance creation.

        Parameters
        ----------
        obj : |Any|
            A class instance with the :attr:`__dict__` attribute.

        """
        vars_dct = vars(obj)
        for k, v in vars_dct.items():
            cls = type(v)
            try:  # More foolproof than .__init__(), especially for empty class instances
                vars_dct[k] = cls.__new__(cls)
            except TypeError:
                # For some classes even the .__new__ method fails (e.g. range or np.ndarray objects)
                vars_dct[k] = None

    """############################# Context manager magic methods #############################"""

    def __enter__(self) -> Tuple[Molecule]:
        """Enter the :class:`SplitMol` context manager; return a list of molecules."""
        # Create a backup of mols' instance variables
        mol = self.mol
        self._vars_backup = vars(mol)
        mol.__dict__ = {k: copy.copy(v) for k, v in vars(mol).items()}

        # Add capping atoms along the to-be split bonds
        self._at_pairs = [self.cap_fragments(bond) for bond in self.bonds]

        # Actually delete the to-be split bonds and split the molecule
        for bond in self.bonds:
            mol.delete_bond(bond)
        self._tmp_mol_list = mol.separate_mod()

        # Lock al instance variables of mol
        # Accessing mol will now raise a MoleculeError for as long as the context manager is open
        self.lock_mol()
        return self._tmp_mol_list

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the :class:`SplitMol` context manager."""
        # Restore all instance variables of mol
        self.unlock_mol()

        # Delete all capping atoms recreate the previously broken bond(s)
        self.reassemble()

        # Delete all instance variables of the temporary molecules
        # All variables are replaced with empty instance of their respective class
        for mol_tmp in self._tmp_mol_list:
            self.reset_vars(mol_tmp)


class FrozenList(list):
    """A list subclass that will raise a |MoleculeError| upon modification of an instance."""

    def __init__(self, *args,
                 exc_type: Type[Exception] = MoleculeError,
                 exc_msg: str = ("'Molecule' objects are inaccessible while opened in "
                                 "the 'SplitMol' context manager"),
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.exc_type = exc_type
        self.exc_msg = exc_msg

    def raise_exception(self, *args, **kwargs) -> NoReturn: raise self.exc_type(self.exc_msg)
    __delitem__ = __iadd__ = __setitem__ = raise_exception
    append = clear = extend = insert = pop = remove = reverse = sort = raise_exception



"""A module designed for sanitizing and interpreting all molecule-related settings in the input file.

Index
-----
.. currentmodule:: CAT.data_handling.validate_mol
.. autosummary::
    santize_smiles
    validate_mol
    _parse_name_type
    _parse_mol_type

API
---
.. autofunction:: santize_smiles
.. autofunction:: validate_mol
.. autofunction:: _parse_name_type
.. autofunction:: _parse_mol_type

"""  # noqa: E501

from os.path import join, isfile, isdir, basename
from typing import Sequence, Any, Union, Optional

from rdkit import Chem
from scm.plams import Settings, Molecule
import scm.plams.interfaces.molecule.rdkit as molkit

from .validation_schemas import mol_schema
from ..utils import validate_path as validate_path_func

__all__ = ['validate_mol', 'santize_smiles']


def santize_smiles(smiles: str) -> str:
    """Sanitize a SMILES string: turn it into a valid filename."""
    name = smiles.replace('(', '[').replace(')', ']')
    try:
        cis_trans = [item for item in smiles if item in ('/', '\\')]
        if cis_trans:
            cis_trans = [item + cis_trans[i*2+1] for i, item in enumerate(cis_trans[::2])]
            cis_trans_dict = {'//': 'trans-', '/\\': 'cis-'}
            for item in cis_trans[::-1]:
                name = cis_trans_dict[item] + name
    except KeyError:
        pass
    finally:
        return name.replace('/', '').replace('\\', '')


def validate_mol(args: Sequence[Union[Any, Settings]],
                 mol_type: str,
                 path: Optional[str] = None,
                 validate_path: bool = True) -> None:
    r"""Validate the ``"input_ligands"``, ``"input_cores"`` and ``"input_qd"`` blocks in the input.

    Performs an inpalce update of **args**.

    Examples
    --------
    An example using a list of .xyz files as input

    .. code:: python

        # A list of .xyz files
        >>> args1 = ['mol1.xyz', 'mol2.xyz']
        >>> mol_type = 'input_ligands'

        >>> validate_mol(args1, mol_type)

        >>> print(args1[0], '\n', args1[1])  # doctest: +SKIP
        is_core:        False
        mol:    /path/to/my/current/working/dir/mol1.xyz
        name:   mol1
        path:   /path/to/my/current/working/dir
        type:   smiles

        is_core:        False
        mol:    /path/to/my/current/working/dir/mol2.xyz
        name:   mol2
        path:   /path/to/my/current/working/dir
        type:   smiles


    Another example using a list of .pdb-containing dictionaries as input

    .. code :: python

        # A list of Settings instances with .xyz files
        >>> print(args2)  # doctest: +SKIP
        [mol3.pdb:
             guess_bonds:       True
         mol4.pdb:
             guess_bonds:       True
        ]

        >>> mol_type = 'input_ligands'
        >>> path = '/path/to/custom/working/dir'
        >>> validate_mol(args2, mol_type, path)  # doctest: +SKIP

        >>> print(args2[0], '\n', args2[1])  # doctest: +SKIP
        guess_bonds:    True
        is_core:        True
        mol:    /path/to/custom/working/dir/mol3.pdb
        name:   mol3
        path:   /path/to/custom/working/dir
        type:   smiles

        guess_bonds:    True
        is_core:        True
        mol:    /path/to/custom/working/dir/mol4.pdb
        name:   mol4
        path:   /path/to/custom/working/dir
        type:   smiles


    Parameters
    ----------
    args : |list|_ [|plams.Settings|_]
        A list of input molecules.
        Accepts strings, PLAMS molecules and RDKit molecules.
        Additional arguments can be provided by putting above-mentioned molecules in a dictionary.

    mol_type : str
        The type of molecule.
        Accepted values are ``"input_ligands"`` and ``"input_cores"``.

    path : str
        Optional: The path to the molecule-containing directory.

    Raises
    ------
    FileNotFoundError
        Raised if **path** cannot be found.

    NotADirectoryError
        Raised if **path** is not a directory.

    ValueError
        Raised if the **mol_type** parameter is neither ``"input_cores"`` nor ``"input_ligands"``.

    SchemaError
        Raised if invalid input settings are found in while validating **args**.

    """
    # Validate arguments
    is_core = _check_core(mol_type)
    is_qd = _check_qd(mol_type)
    if validate_path:
        _path = validate_path_func(path)
    else:
        _path = path

    for i, dict_ in enumerate(args):
        if not isinstance(dict_, dict):  # No optional arguments provided
            mol = dict_
            mol_dict = Settings({'path': _path, 'is_core': is_core, 'is_qd': is_qd})
        else:  # Optional arguments have been provided: parse and validate them
            if dict_.get('parsed'):
                continue
            mol, mol_dict = next(iter(dict_.items()))
            mol_dict.setdefault('is_core', is_core)
            mol_dict.setdefault('is_qd', is_qd)
            mol_dict = mol_schema.validate(mol_dict)
            mol_dict.setdefault('path', _path)

        if isinstance(mol, str) and not isfile(mol):
            mol = join(mol_dict.path, mol)
        mol_dict.mol = mol
        mol_dict.parsed = True

        _parse_name_type(mol_dict)
        args[i] = mol_dict


def _parse_name_type(mol_dict: Settings) -> None:
    """Set the ``"name"`` and ``"type"`` keys in **mol_dict**.

    The new values of ``"name"`` and ``"type"`` depend on the value of ``mol_dict["mol"]``.

    Parameters
    ----------
    mol_dict : |plams.Settings|_
        A Settings instance containing the ``"mol"`` key.
        ``mol_dict["mol"]`` is exp

    Raises
    ------
    TypeError
        Raised ``mol_dict["mol"]`` is an instance of neither :class:`str`, :class:`Molecule` nor
        :class:`mol`.

    """
    mol = mol_dict.mol
    if isinstance(mol, str):
        if isfile(mol):  # mol is a file
            mol_dict.type = mol.rsplit('.', 1)[-1]
            mol_dict.name = basename(mol.rsplit('.', 1)[0])
        elif isdir(mol):  # mol is a directory
            mol_dict.type = 'folder'
            mol_dict.name = basename(mol)
        else:  # mol is (probably; hopefully?) a SMILES string
            i = 1 + len(mol_dict.path) if 'path' in mol_dict else 0
            mol_dict.type = 'smiles'
            mol_dict.mol = mol[i:]
            mol_dict.name = santize_smiles(mol_dict.mol)

    elif isinstance(mol, Molecule):  # mol is an instance of plams.Molecule
        mol_dict.type = 'plams_mol'
        if not mol.properties.name:
            mol_dict.name = Chem.MolToSmiles(Chem.RemoveHs(molkit.to_rdmol(mol)), canonical=True)
        else:
            mol_dict.name = mol.properties.name

    elif isinstance(mol, Chem.rdchem.Mol):  # mol is an instance of rdkit.Chem.Mol
        mol_dict.type = 'rdmol'
        mol_dict.name = Chem.MolToSmiles(Chem.RemoveHs(mol), canonical=True)

    else:
        raise TypeError(f"mol_dict['mol'] expects an instance of 'str', 'Molecule' or 'Mol'; "
                        f"observed type: '{mol.__class__.__name__}'")


def _check_core(mol_type: str) -> bool:
    """Check the **mol_type** parameter of :func:`.validate_mol`."""
    if mol_type.lower() == 'input_cores':
        return True
    elif mol_type.lower() in ('input_ligands', 'input_qd'):
        return False
    else:
        raise ValueError(f"accepted values for mol_type are 'input_cores', 'input_ligands' and "
                         f"'input_qd'; observed value: {repr(mol_type)}")


def _check_qd(mol_type: str) -> bool:
    """Parse the **mol_type** parameter of :func:`.validate_mol`."""
    if mol_type.lower() == 'input_qd':
        return True
    elif mol_type.lower() in ('input_ligands', 'input_cores'):
        return False
    else:
        raise ValueError(f"accepted values for mol_type are 'input_qd' and input_ligands; "
                         f"observed value: {repr(mol_type)}")



"""A module designed for finding core functional groups.

Index
-----
.. currentmodule:: CAT.attachment.ligand_anchoring
.. autosummary::
    set_core_anchors
    find_core_substructure

API
---
.. autofunction:: set_core_anchors
.. autofunction:: find_core_substructure

"""

from typing import Tuple, Any, Mapping, TYPE_CHECKING

import numpy as np
from scm.plams import Molecule, MoleculeError, to_rdmol

from .distribution import distribute_idx
from ..utils import AllignmentEnum, AllignmentTup, AnchorTup

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from numpy import int64 as i8

__all__ = ["set_core_anchors", "find_core_substructure"]


def set_core_anchors(
    mol: Molecule,
    anchor_tup: AnchorTup,
    allignment_tup: AllignmentTup,
    subset_kwargs: "None | Mapping[str, Any]" = None,
) -> Tuple[str, str]:
    """Identify and parse the core anchors within the passed molecule.

    Returns two strings: The (parsed) molecular formula and the anchor indices.

    """
    # Checks the if the anchor is a string (atomic symbol) or integer (atomic number)
    formula = mol.get_formula()

    # Get the indices of all anchor atom ligand placeholders in the core
    anchors = mol.properties.dummies
    if not anchors:
        anchor_idx, remove_idx = find_core_substructure(mol, anchor_tup)
    else:
        anchor_idx = np.fromiter(anchors, count=len(anchors), dtype=np.int64)
        anchor_idx -= 1
        remove_idx = anchor_idx.copy()
    if subset_kwargs:
        anchor_idx = distribute_idx(mol, anchor_idx, **subset_kwargs)
    if not len(anchor_idx):
        raise MoleculeError(f"No valid anchoring groups found in the core {formula!r}")

    # Convert atomic indices into Atoms
    anchor_idx += 1
    anchor_idx.sort()
    mol.properties.dummies = [mol[i] for i in anchor_idx]

    # Returns an error if no anchor atoms were found
    if (len(mol) - len(anchor_idx)) < 4 and allignment_tup.kind == AllignmentEnum.SURFACE:
        raise NotImplementedError(
            '`optional.core.allignment = "surface"` is not supported for cores with less '
            f'than 4 (non-anchor) atoms ({mol.get_formula()}); consider using '
            '`optional.core.allignment = "sphere"`'
        )

    # Delete all core anchor atoms
    if remove_idx is not None:
        remove_idx += 1
        remove_idx.sort()
        for i in reversed(remove_idx):
            mol.delete_atom(mol[i])
    return formula, ' '.join(anchor_idx.astype(str))


def find_core_substructure(
    mol: Molecule,
    anchor_tup: AnchorTup,
) -> Tuple["NDArray[i8]", "None | NDArray[i8]"]:
    """Identify substructures within the passed core based on **anchor_tup**.

    Returns two indice-arrays, respectivelly containing the indices of the anchor
    atoms and all to-be removed atoms.

    """
    rdmol = to_rdmol(mol)
    matches = rdmol.GetSubstructMatches(anchor_tup.mol, useChirality=True)
    remove = anchor_tup.remove

    # Remove all duplicate matches, each heteroatom (match[0]) should have <= 1 entry
    ref_set = set()
    anchor_list = []
    remove_list = []
    for idx_tup in matches:
        anchor_idx_tup = tuple(idx_tup[i] for i in anchor_tup.group_idx)
        if anchor_idx_tup in ref_set:
            continue  # Skip duplicates
        else:
            ref_set.add(anchor_idx_tup)

        if remove is not None:
            remove_list += [idx_tup[i] for i in remove]
        anchor_list.append(anchor_idx_tup[0])

    anchor_array = np.fromiter(anchor_list, dtype=np.int64, count=len(anchor_list))
    if remove is not None:
        remove_array = np.fromiter(remove_list, dtype=np.int64, count=len(remove_list))
        return anchor_array, remove_array
    else:
        return anchor_array, None



"""A module for constructing vectors perpendicular to a molecules surface.

Index
-----
.. currentmodule:: CAT.attachment.perp_surface
.. autosummary::
    get_surface_vec
    plot_vectors

API
---
.. autofunction:: get_surface_vec
.. autofunction:: plot_vectors

"""

from typing import Union, Optional, Any

import numpy as np
from scipy.spatial import ConvexHull, cKDTree

from scm.plams import Molecule

try:
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

    PLT: Optional[ImportError] = None
    Figure = plt.Figure
except ImportError as ex:
    PLT = ex
    Figure = 'matplotlib.pyplot.Figure'

__all__ = ['get_surface_vec', 'plot_vectors']


def get_surface_vec(mol: Union[Molecule, np.ndarray],
                    anchor: Union[None, Molecule, np.ndarray] = None) -> np.ndarray:
    """Construct a set of vectors perpendicular to the surface of **mol** and assign one to each atom in **anchor**.

    Utilizes a convex hull algorithm for identifying and partitioning the surface.

    Parameters
    ----------
    mol : array-like [:class:`float`], shape :math:`(n, 3)`
        A 2D array-like object with the Cartesian coordinates of a molecule.

    anchor : array-like [:class:`float`], shape :math:`(m, 3)`, optional
        A 2D array-like object with the Cartesian coordinates of a set of anchor atoms.
        If ``None``, default to **mol**.

    Returns
    -------
    :class:`numpy.ndarray` [:class:`float`], shape :math:`(m, 3)`
        An array with vectors perpendicular to the surface of **mol**,
        one for each atom in **anchor**.

    See Also
    --------
    :class:`ConvexHull<scipy.spatial.ConvexHull>`
        Convex hulls in N dimensions.

    """  # noqa
    xyz = np.array(mol, dtype=float, ndmin=2, copy=False)
    if anchor is None:
        anchor = xyz
    else:
        anchor = np.array(anchor, dtype=float, ndmin=2, copy=False)

    # Construct the convex hull and extract the vertices
    hull = ConvexHull(xyz)
    simplice = np.swapaxes(xyz[hull.simplices], 0, 1)
    simplice_center = simplice.mean(axis=0)

    # Construct and return the surface vectors
    vec = _get_perp_vecs(*simplice)
    _flip_vec(simplice_center, vec)
    return vec[_find_nearest_center(anchor, simplice_center)]


def plot_vectors(vec: np.ndarray,
                 xyz: Optional[np.ndarray] = None,
                 show: bool = True, **kwargs: Any) -> Figure:
    r"""Create a 3D plot of all (3D) vectors in **vec**.

    Parameters
    ----------
    vec : array-like [:class:`float`], shape :math:`(n, 3)`
        A 2D array-like object representing :math:`n` vectors.

    xyz : array-like [:class:`float`], shape :math:`(n, 3)`, optional
        An array with the Cartesian coordinates defining the
        origin of each vector in **vec**.
        If ``None``, default to the origin (:code:`[0, 0, 0]`).

    show : :class:`bool`
        Show the created figure.

    \**kwargs : :data:`Any<typing.Any>`
        Further keyword arguments for
        :meth:`Axes.quiver()<matplotlib.pyplot.Axes.quiver>`
        such as the **length** keyword.

    Returns
    -------
    :class:`Figure<matplotlib.pyplot.Figure>`
        The resulting matplotlib Figure.

    """
    if PLT is not None:
        raise PLT

    # Parse arguments
    vec = np.array(vec, ndmin=2, dtype=float, copy=False)
    if xyz is not None:
        xyz = np.array(xyz, ndmin=2, dtype=float, copy=False)
    else:
        xyz = np.array(0.0, ndmin=2)

    # Extract the x, y and z coordinates
    if xyz.size == 1:
        x = y = z = xyz
    else:
        x, y, z = xyz.T

    # Extract the x, y and z components of the vector
    u, v, w = vec.T

    # Construct the figure
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    _set_axis_limit(ax, x, y, z, u, v, w)
    ax.quiver(x, y, z, u, v, w, **kwargs)

    if show:
        plt.show(block=True)
    return fig


def _set_axis_limit(ax, x, y, z, u, v, w) -> None:
    """Set the axis limits for :func:`plot_vectors`."""
    x_mean = x.mean()
    y_mean = y.mean()
    z_mean = z.mean()
    dx = max(1, max(max(abs(x)), max(abs(u))) - x_mean)
    dy = max(1, max(max(abs(y)), max(abs(v))) - y_mean)
    dz = max(1, max(max(abs(z)), max(abs(w))) - z_mean)
    ax.set_xlim([x_mean-dx, x_mean+dx])
    ax.set_ylim([y_mean-dy, y_mean+dy])
    ax.set_zlim([z_mean-dz, z_mean+dz])


def _find_nearest_center(anchor: np.ndarray, center: np.ndarray) -> np.ndarray:
    """Find the points in **center** closest to those in **anchor**."""
    tree = cKDTree(center)
    _, idx = tree.query(anchor, k=1)
    return idx


def _get_perp_vecs(p1: np.ndarray, p2: np.ndarray, p3: np.ndarray) -> np.ndarray:
    """Construct a unit vector perpendicular to a set of triangular polygons."""
    v1 = p2 - p1
    v2 = p3 - p1
    vec = np.cross(v1, v2)
    vec /= np.linalg.norm(vec, axis=-1)[..., None]
    return vec


def _flip_vec(simplice_centra: np.ndarray, perp_vec: np.ndarray) -> None:
    """Ensure that **perp_vec** is pointing away from the surface defined by **simplice_centra**."""
    center = simplice_centra.mean(axis=0)
    dot = np.einsum('ij,ij->i', simplice_centra - center, perp_vec)
    perp_vec[dot < 0] *= -1



"""Utility functions related to the :mode:`distribution<CAT.attachment.distribution>` module.

Index
-----
.. currentmodule:: CAT.attachment.distribution_utils
.. autosummary::
    test_distribute

API
---
.. autofunction:: test_distribute

"""

from typing import Any, Union, Optional, Tuple, Iterable
from collections import abc

import numpy as np
from scipy.spatial.distance import cdist

from scm.plams import Molecule, Atom, rotation_matrix

from .distribution import distribute_idx


def test_distribute(mol: Union[Molecule, str], symbol: str,
                    f_range: Union[float, Iterable[float]],
                    rotate: Optional[Tuple[float, float, float]] = (0.1, -0.1, 0.9),
                    **kwargs: Any) -> Molecule:
    r"""Test function for :func:`CAT.attachment.distribution.distribute_idx`.

    Examples
    --------
    .. code:: python

        >>> import numpy as np
        >>> from scm.plams import Molecule

        >>> mol_input = Molecule()
        >>> xyz_output: str = ...
        >>> at_symbol = 'Cl'
        >>> f_range: np.ndarray = 2**-np.arange(8.0)

        >>> mol_out: Molecule = test_distribute(mol_input, at_symbol, f_range)  # doctest: +SKIP
        >>> mol_out.write(xyz_output)  # doctest: +SKIP

        >>> print(len(mol_input) == len(p_range) * len(mol_out))  # doctest: +SKIP
        True

    Parameters
    ----------
    mol : :class:`Molecule` or :class:`str`
        A molecule or path+filename containing a molecule.

    symbol : :class:`str`
        The atomic symbol of the anchor atom.

    f_range : :class:`float` or :class:`Iterable<collections.abc.Iterable>` [:class:`float`]
        A float or iterable of floats subject to the following constraint: :math:`0 < f \le 1`.

    rotate : :class:`Sequence<collection.abc.Sequence>` [:class:`float`], shape :math:`(3,)`, optional
        A sequence of three floats representing a molecular orientation.

    \**kwargs : :data:`Any<typing.Any>`
        Further keyword arguments for
        :func:`distribute_idx()<CAT.attachment.distribution.distribute_idx>`:
        ``follow_edge``, ``mode`` and ``start``.

    Returns
    -------
    :class:`Molecule`
        A Molecule instance containing one molecule for every item in **p_range**

    """  # noqa
    if not isinstance(mol, Molecule):
        mol = Molecule(mol)
    if not isinstance(f_range, abc.Iterable):
        f_range = (f_range,)

    ret = Molecule()
    trans = cdist(mol, mol).max() * 1.1
    for i, f in enumerate(f_range):
        mol_tmp = _test_distribute(mol, symbol, f=f, **kwargs)
        if rotate is not None:
            mol_tmp.rotate(rotation_matrix([0, 0, 1], rotate))
        mol_tmp.translate([i*trans, 0, 0])
        ret += mol_tmp
    return ret


def _test_distribute(mol: Molecule, symbol: str, **kwargs) -> Molecule:
    """Helper function for :func:`test_distribute`."""
    if not isinstance(mol, Molecule):
        mol = Molecule(mol)

    _idx_in = [i for i, at in enumerate(mol) if at.symbol == symbol]
    idx_in = np.fromiter(_idx_in, count=len(_idx_in), dtype=int)
    idx_out = distribute_idx(mol, idx_in, **kwargs)

    a = symbol
    b = 'I' if a != 'I' else 'Br'
    mol2 = Molecule()
    for i, at in enumerate(mol):
        if at.symbol != symbol:
            continue
        symbol_new = a if i not in idx_out else b
        mol2.add_atom(Atom(symbol=symbol_new, coords=at.coords, mol=mol2))
    return mol2



"""A module for updating the index of quantum dot dataframes.

Index
-----
.. currentmodule:: CAT.data_handling.update_qd_df
.. autosummary::
    update_qd_df

API
---
.. autofunction:: update_qd_df

"""

import pandas as pd

from scm.plams import Molecule

from ..workflows import MOL, OPT, HDF5_INDEX

__all__ = ['update_qd_df']


def update_qd_df(qd_df: pd.DataFrame) -> None:
    """Update the index of **qd_df** with descriptors of the core(s) and ligand(s)."""
    # Construct the MultiIndex columns
    core = [_get_core_formula(mol) for mol in qd_df[MOL]]
    core_anchor = [None] * len(qd_df)
    ligand_smiles = [mol.properties.pop('ligand_smiles') for mol in qd_df[MOL]]
    ligand_anchor = [mol.properties.pop('ligand_anchor') for mol in qd_df[MOL]]

    # Reassign the index
    idx_iter = zip(*(core, core_anchor, ligand_smiles, ligand_anchor))
    names = ['core', 'core anchor', 'ligand smiles', 'ligand anchor']
    qd_df.index = pd.MultiIndex.from_tuples(idx_iter, names=names)

    # Set the ('opt', '') column to False
    qd_df[OPT] = False
    qd_df[HDF5_INDEX] = -1


def _get_core_formula(mol: Molecule) -> str:
    """Construct a string representing the molecular formula of the core."""
    # Create a dictionary with atomic symbols and matching atom counts
    core = {}
    for i in mol.properties.indices:
        at = mol[i]
        if 'anchor' not in at.properties:
            break  # The first ligand has been reached; abort and return
        symbol = at.symbol
        try:
            core[symbol] += 1
        except KeyError:
            core[symbol] = 1

    # Concatenate the dictionary into a single string
    return ''.join(f'{k}{v}' for k, v in sorted(core.items()))



"""A number of recipes constructed using the CAT and Nano-CAT packages.

Examples
--------
.. code:: python

    >>> from CAT.recipes import bulk_workflow  # doctest: +SKIP
    >>> from CAT.recipes import replace_surface  # doctest: +SKIP
    >>> from CAT.recipes import dissociate_surface, row_accumulator  # doctest: +SKIP
    >>> from CAT.recipes import coordination_number  # doctest: +SKIP
    >>> from CAT.recipes import add_ligands, export_dyes, sa_scores
    ...

"""

# flake8: noqa: F403,F401

try:
    from nanoCAT import recipes as _recipes
    from nanoCAT.recipes import *

    __all__ = _recipes.__all__.copy()
    del _recipes

except ImportError as ex:
    import warnings as _warnings

    __all__ = []

    _warning = ImportWarning(str(ex))
    _warning.__cause__ = ex
    _warnings.warn(_warning)
    del _warning
    del _warnings

finally:
    from . import dye as _dye
    from .dye import *

    __all__ += _dye.__all__
    del _dye



"""The CAT logger.

Index
-----
.. currentmodule:: CAT.logger
.. autosummary::
    logger

API
---
.. autodata:: logger
    :annotation: = logging.RootLogger

"""

import logging

from scm.plams import Molecule
from scm.plams.core.basejob import Job

__all__ = ['logger']

#: The CAT logger
logger = logging.getLogger('CAT')
_handler = logging.StreamHandler()
_handler.setLevel(logging.DEBUG)
_handler.setFormatter(logging.Formatter(
    fmt='[%(asctime)s] %(levelname)s: %(message)s',
    datefmt='%H:%M:%S',
))
logger.addHandler(_handler)
del _handler


def log_start(job: Job, mol: Molecule, job_preset: str, job_name: str) -> None:
    """Log the start of a PLAMS :class:`Job`."""
    msg = 'has started'
    args = job.__class__.__name__, mol.properties.name, job_preset, job_name, msg
    logger.info('{}: {} {} ({}) {}'.format(*args))


def log_succes(job: Job, mol: Molecule, job_preset: str, job_name: str) -> None:
    """Log the succesful termination of a PLAMS :class:`Job`."""
    msg = 'is successful'
    args = job.__class__.__name__, mol.properties.name, job_preset, job_name, msg
    logger.info('{}: {} {} ({}) {}'.format(*args))


def log_fail(job: Job, mol: Molecule, job_preset: str, job_name: str) -> None:
    """Log the unsuccesful termination of a PLAMS :class:`Job`."""
    msg = 'has failed'
    args = job.__class__.__name__, mol.properties.name, job_preset, job_name, msg
    logger.info('{}: {} {} ({}) {}'.format(*args))


def log_copy(job: Job, mol: Molecule, job_preset: str, job_name: str, job_old: Job) -> None:
    """Log the copying of a previous PLAMS :class:`Job`."""
    msg = f'results copied from {job_old.name}'
    args = job.__class__.__name__, mol.properties.name, job_preset, job_name, msg
    logger.info('{}: {} {} ({}) {}'.format(*args))



"""A module related to the importing of molecules.

Index
-----
.. currentmodule:: CAT.data_handling.mol_import
.. autosummary::
    read_mol
    read_mol_xyz
    read_mol_pdb
    read_mol_mol
    read_mol_smiles
    read_mol_plams
    read_mol_rdkit
    read_mol_folder
    read_mol_txt
    get_charge_dict
    set_mol_prop
    set_atom_prop
    set_qd
    print_exception

API
---
.. autofunction:: read_mol
.. autofunction:: read_mol_xyz
.. autofunction:: read_mol_pdb
.. autofunction:: read_mol_mol
.. autofunction:: read_mol_smiles
.. autofunction:: read_mol_plams
.. autofunction:: read_mol_rdkit
.. autofunction:: read_mol_folder
.. autofunction:: read_mol_txt
.. autofunction:: get_charge_dict
.. autofunction:: set_mol_prop
.. autofunction:: set_atom_prop
.. autofunction:: set_qd
.. autofunction:: print_exception

"""

import os
import itertools
from types import MappingProxyType
from string import ascii_letters
from typing import Dict, Iterable, List, Sequence, Optional, Mapping, Callable, overload

import numpy as np

from scm.plams import (Molecule, Atom, Settings)
import scm.plams.interfaces.molecule.rdkit as molkit

from rdkit import Chem, RDLogger

from ..utils import get_formula
from ..logger import logger
from ..data_handling.validate_mol import validate_mol

__all__ = ['read_mol', 'set_mol_prop']

# Supress the rdkit logger
_logger = RDLogger.logger()
_logger.setLevel(RDLogger.CRITICAL)


@overload
def read_mol(input_mol: Iterable[Settings]) -> List[Molecule]:
    ...
@overload  # noqa: E302
def read_mol(input_mol: None) -> None:
    ...
def read_mol(input_mol):  # noqa: E302
    """Checks the filetypes of the input molecules.

    Sets the molecules' properties and returns a list of plams molecules.

    Parameters
    ----------
    input_mol : |list|_ [|Settings|_]
        An iterable consisting of dictionaries with input settings per mol.

    Returns
    -------
    |plams.Molecule|_
        A list of plams Molecules.

    """
    if input_mol is None:
        return None

    # Create a list of PLAMS molecules
    mol_list = []
    append = mol_list.append
    extend = mol_list.extend

    for mol_dict in input_mol:
        try:
            read_mol = EXTENSION_MAPPING[mol_dict.type]
        except KeyError as ex:
            logger.error(f'{ex.__class__.__name__}: {ex}')
            continue

        mol = read_mol(mol_dict)
        if not mol:  # Failed to import any molecules
            continue

        if isinstance(mol, list):  # if mol is a list of molecules
            extend(mol)

        elif mol_dict.is_qd:  # A quantum dot molecule
            set_qd(mol, mol_dict)
            append(mol)

        else:  # A core or ligand molecule
            if mol_dict.guess_bonds:
                mol.guess_bonds()
            set_mol_prop(mol, mol_dict)
            append(mol)

    return mol_list


def read_mol_xyz(mol_dict: Settings) -> Optional[Molecule]:
    """Read an .xyz file."""
    try:
        mol = Molecule(mol_dict.mol, inputformat='xyz')
        if mol_dict.guess_bonds and not mol_dict.is_qd:
            mol.guess_bonds()
        if not mol_dict.is_core:
            canonicalize_mol(mol)
        return mol
    except Exception as ex:
        print_exception('read_mol_xyz', mol_dict.name, ex)


def read_mol_pdb(mol_dict: Settings) -> Optional[Molecule]:
    """Read a .pdb file."""
    try:
        mol = molkit.readpdb(mol_dict.mol)
        if mol_dict.guess_bonds and not mol_dict.is_qd:
            mol.guess_bonds()
        if not mol_dict.is_core:
            canonicalize_mol(mol)
        return mol
    except Exception as ex:
        print_exception('read_mol_pdb', mol_dict.name, ex)


def read_mol_mol(mol_dict: Settings) -> Optional[Molecule]:
    """Read a .mol file."""
    try:
        mol = molkit.from_rdmol(Chem.MolFromMolFile(mol_dict.mol, removeHs=False))
        if mol_dict.guess_bonds and not mol_dict.is_qd:
            mol.guess_bonds()
        if not mol_dict.is_core:
            canonicalize_mol(mol)
        return mol
    except Exception as ex:
        print_exception('read_mol_mol', mol_dict.name, ex)


def read_mol_smiles(mol_dict: Settings) -> Optional[Molecule]:
    """Read a SMILES string."""
    try:
        mol = _from_smiles(mol_dict.mol)
        mol.properties.smiles = Chem.CanonSmiles(mol_dict.mol)

        if mol_dict.get('indices'):
            for i in mol_dict.indices:
                mol[i].properties.anchor = True

        canonicalize_mol(mol)
        if mol_dict.get('indices'):
            mol_dict.indices = tuple(i for i, at in enumerate(mol, 1) if
                                     at.properties.pop('anchor', False))

        if mol_dict.guess_bonds and not mol_dict.is_qd:
            mol.guess_bonds()
        return mol
    except Exception as ex:
        print_exception('read_mol_smiles', mol_dict.name, ex)


def _from_smiles(smiles: str) -> Molecule:
    sanitize = Chem.SanitizeFlags.SANITIZE_ALL ^ Chem.SanitizeFlags.SANITIZE_ADJUSTHS
    _rdmol = Chem.MolFromSmiles(smiles, sanitize=False)
    Chem.rdmolops.SanitizeMol(_rdmol, sanitizeOps=sanitize)

    rdmol = Chem.AddHs(_rdmol)
    rdmol.SetProp('smiles', smiles)
    return molkit.get_conformations(rdmol, 1, None, None, rms=0.1)


def read_mol_plams(mol_dict: Settings) -> Optional[Molecule]:
    """Read a PLAMS molecule."""
    try:
        mol = mol_dict.mol
        if mol_dict.guess_bonds and not mol_dict.is_qd:
            mol.guess_bonds()
        if not mol_dict.is_core:
            canonicalize_mol(mol)
        return mol
    except Exception as ex:
        print_exception('read_mol_plams', mol_dict.name, ex)


def read_mol_rdkit(mol_dict: Settings) -> Optional[Molecule]:
    """Read a RDKit molecule."""
    try:
        mol = molkit.from_rdmol(mol_dict.mol)
        if mol_dict.guess_bonds and not mol_dict.is_qd:
            mol.guess_bonds()
        if not mol_dict.is_core:
            canonicalize_mol(mol)
        return mol
    except Exception as ex:
        print_exception('read_mol_rdkit', mol_dict.name, ex)


def read_mol_folder(mol_dict: Settings) -> Optional[List[Molecule]]:
    """Read all files (.xyz, .pdb, .mol, .txt or further subfolders) within a folder."""
    try:
        mol_type = 'input_cores' if mol_dict.is_core else 'input_ligands'

        _file_list = os.listdir(mol_dict.mol)
        optional_dict = Settings({k: v for k, v in mol_dict.items() if k not in ('mol', 'path')})
        file_list = [{i: optional_dict} for i in _file_list]

        validate_mol(file_list, mol_type, mol_dict.path)
        return read_mol(file_list)
    except Exception as ex:
        print_exception('read_mol_folder', mol_dict.name, ex)


def read_mol_txt(mol_dict: Settings) -> Optional[List[Molecule]]:
    """Read a plain text file containing one or more SMILES strings."""
    try:
        row = 0 if 'row' not in mol_dict else mol_dict.row
        column = 0 if 'column' not in mol_dict else mol_dict.column
        mol_type = 'input_cores' if mol_dict.is_core else 'input_ligands'

        with open(mol_dict.mol, 'r') as f:
            iterator = itertools.islice(f, row, None)
            _file_list = [i.rstrip('\n').split()[column] for i in iterator if i]
        optional_dict = Settings({k: v for k, v in mol_dict.items() if k not in ('mol', 'path')})
        file_list = [{i: optional_dict} for i in _file_list]

        validate_mol(file_list, mol_type, mol_dict.path)
        return read_mol(file_list)
    except Exception as ex:
        print_exception('read_mol_txt', mol_dict.name, ex)


EXTENSION_MAPPING: Mapping[str, Callable[[Settings], Optional[Molecule]]] = MappingProxyType({
    'xyz': read_mol_xyz,
    'pdb': read_mol_pdb,
    'mol': read_mol_mol,
    'smiles': read_mol_smiles,
    'folder': read_mol_folder,
    'txt': read_mol_txt,
    'plams_mol': read_mol_plams,
    'rdmol': read_mol_rdkit
})


def canonicalize_mol(mol: Molecule, inplace: bool = True) -> Optional[Molecule]:
    """Take a PLAMS molecule and sort its atoms based on their canonical rank.

    .. _rdkit.Chem.CanonicalRankAtoms: https://www.rdkit.org/docs/source/rdkit.Chem.rdmolfiles.html#rdkit.Chem.rdmolfiles.CanonicalRankAtoms

    Examples
    --------
    .. code:: python

        >>> from scm.plams import Molecule, from_smiles

        # Methane
        >>> mol: Molecule = from_smiles('C')
        >>> print(mol)  # doctest: +SKIP
        Atoms:
            1         H      0.640510      0.640510     -0.640510
            2         H      0.640510     -0.640510      0.640510
            3         C      0.000000      0.000000      0.000000
            4         H     -0.640510      0.640510      0.640510
            5         H     -0.640510     -0.640510     -0.640510

        >>> canonicalize_mol(mol)
        >>> print(mol)  # doctest: +SKIP
        Atoms:
            1         C      0.000000      0.000000      0.000000
            2         H     -0.640510     -0.640510     -0.640510
            3         H     -0.640510      0.640510      0.640510
            4         H      0.640510     -0.640510      0.640510
            5         H      0.640510      0.640510     -0.640510

    Parameters
    ----------
    mol : |plams.Molecule|_
        A PLAMS molecule.

    inplace : bool
        If ``True``, perform an inplace update of **mol** rather than returning
        a new :class:`Molecule` instance.

    Returns
    -------
    |plams.Molecule|_
        Optional: if ``inplace=False``, return a copy of **mol** with its atoms sorted by their
        canonical rank.

    See Also
    --------
    * rdkit.Chem.CanonicalRankAtoms_: Returns the canonical atom ranking for each atom of a
      molecule fragment.

    """  # noqa
    rdmol = molkit.to_rdmol(mol)
    idx_collection = Chem.CanonicalRankAtoms(rdmol)

    # Reverse sort Molecule.atoms by the atomic indices in idx_collection
    if inplace:
        mol.atoms = [at for _, at in sorted(zip(idx_collection, mol.atoms), reverse=True)]
        return
    else:
        ret = mol.copy()
        ret.atoms = [at for _, at in sorted(zip(idx_collection, ret.atoms), reverse=True)]
        return ret


def get_charge_dict() -> Dict[str, int]:
    """Create a dictionary of elements and their formal atomic charge."""
    # Create a list of atomic charges and elements
    charges = (1, 2, -3, -2, -1, 2)
    elements = (
        ('Li', 'Na', 'K', 'Rb', 'Cs'),      # Group 1: Alkali metals
        ('Be', 'Mg', 'Ca', 'Sr', 'Ba'),     # Group 2: Alkaline earth metals
        ('N', 'P', 'As', 'Sb', 'Bi'),       # Group 15: pnictogens
        ('O', 'S', 'Se', 'Te', 'Po'),       # Group 16: Chalcogens
        ('H', 'F', 'Cl', 'Br', 'I', 'At'),  # Group 17: Halogens
        ('Cd', 'Pb')                        # Misc
    )

    # Combine the elements and atomic charges into a dictionary
    values = itertools.chain.from_iterable([i] * len(j) for i, j in zip(charges, elements))
    keys = itertools.chain(*elements)

    return dict(zip(keys, values))


#: A dictionary with default atomic charges
CHARGE_MAPPING: Mapping[str, int] = MappingProxyType(get_charge_dict())


def set_mol_prop(mol: Molecule, mol_dict: Settings) -> None:
    """Set molecular and atomic properties."""
    if mol_dict.is_core:
        residue_name = 'COR'
        mol.properties.name = get_formula(mol)
    else:
        residue_name = 'LIG'
        mol.properties.name = mol_dict.name

    mol.properties.dummies = mol_dict.indices
    mol.properties.path = mol_dict.path
    mol.properties.job_path = []

    # Prepare a generator of letters for pdb_info.Name
    alphabet = itertools.combinations(ascii_letters, 2)

    # Set the atomic properties
    for atom, i in zip(mol, itertools.cycle(alphabet)):
        set_atom_prop(atom, i, residue_name)

    if not mol.properties.smiles:
        mol.properties.smiles = Chem.MolToSmiles(
            Chem.RemoveHs(molkit.to_rdmol(mol)), canonical=True
        )


def set_atom_prop(atom: Atom, at_id: Sequence[str], residue_name: str) -> None:
    """Set atomic properties."""
    symbol = '{:4}'.format(atom.symbol + ''.join(at_id))

    # Add a number of properties to atom
    atom.properties.pdb_info.ResidueName = residue_name
    atom.properties.pdb_info.Occupancy = 1.0
    atom.properties.pdb_info.TempFactor = 0.0
    atom.properties.pdb_info.ResidueNumber = 1
    atom.properties.pdb_info.Name = symbol
    atom.properties.pdb_info.ChainId = 'A'

    # Changes hydrogen and carbon from heteroatom to atom
    if atom.symbol in ('H', 'C'):
        atom.properties.pdb_info.IsHeteroAtom = False
    else:
        atom.properties.pdb_info.IsHeteroAtom = True

    # Sets the formal atomic charge
    if atom.properties.charge:
        return

    # Default to a charge of 0 if no charge is available for that specific element
    if atom.symbol not in CHARGE_MAPPING:
        atom.properties.charge = 0
        return

    # Update the charge of non-hypervalent atoms
    total_bonds = int(sum([bond.order for bond in atom.bonds]))
    default_charge = CHARGE_MAPPING[atom.symbol]
    abs_charge = abs(default_charge)
    sign = -1 * int(default_charge / abs_charge)

    # Take the default charge and correct for the number (and order) of bonds
    atom.properties.charge = default_charge + sign * total_bonds
    if total_bonds <= abs_charge:
        return

    # Update formal atomic charges for hypervalent atoms
    if total_bonds is 2 + abs_charge:
        atom.properties.charge -= 2 * sign
    elif total_bonds is 4 + abs_charge:
        atom.properties.charge -= 4 * sign
    elif total_bonds >= 6 + abs_charge:
        atom.properties.charge -= 6 * sign
    return


def set_qd(qd: Molecule, mol_dict: Settings) -> Molecule:
    """Update quantum dots imported by :func:`.read_mol`."""
    # Create ligand (and anchor) molecules
    ligand = molkit.from_smiles(mol_dict.ligand_smiles)
    ligand_rdmol = molkit.to_rdmol(ligand)
    anchor = molkit.from_smiles(mol_dict.ligand_anchor)
    anchor_rdmol = molkit.to_rdmol(anchor)
    qd_rdmol = molkit.to_rdmol(qd)

    # Create arrays of atomic indices of the core and ligands
    lig_idx = 1 + np.array(qd_rdmol.GetSubstructMatches(ligand_rdmol))
    core_idx = np.arange(1, len(qd))[~lig_idx]
    lig_idx = lig_idx.ravel().tolist()
    core_idx = core_idx.tolist()

    # Guess bonds
    if mol_dict.guess_bonds:
        qd.guess_bonds(atom_subset=[qd[i] for i in lig_idx])

    # Reorder all atoms: core atoms first followed by ligands
    qd.atoms = [qd[i] for i in core_idx] + [qd[j] for i in lig_idx for j in i]

    # Construct a list with the indices of all ligand anchor atoms
    core_idx_max = 1 + len(core_idx)
    _anchor_idx = ligand_rdmol.GetSubstructMatch(anchor_rdmol)[0]
    start = core_idx_max + _anchor_idx
    stop = core_idx_max + _anchor_idx + np.product(lig_idx.shape)
    step = len(ligand)
    anchor_idx = list(range(start, stop, step))

    # Update the properties of **qd**
    for i in anchor_idx:
        qd[i].properties.anchor = True
    qd.properties.indices = list(range(1, core_idx_max)) + anchor_idx
    qd.properties.job_path = []
    qd.properties.name = mol_dict.name
    qd.properties.path = mol_dict.path
    qd.properties.ligand_smiles = Chem.CanonSmiles(mol_dict.ligand_smiles)
    qd.properties.ligand_anchor = f'{ligand[_anchor_idx].symbol}{_anchor_idx}'

    # Update the pdb_info of all atoms
    for i, at in enumerate(qd, 1):
        at.properties.pdb_info.SerialNumber = i
        if i <= core_idx_max:  # A core atom
            at.properties.pdb_info.ResidueNumber = 1
        else:  # A ligand atom
            at.properties.pdb_info.ResidueNumber = 2 + int((i - core_idx_max) / len(ligand))


def print_exception(func_name: str, mol_name: str, ex: Exception) -> None:
    """Manages the printing of exceptions upon failing to import a molecule."""
    ex_name = ex.__class__.__name__
    err = f'CAT.{func_name}() failed to parse {repr(mol_name)}; Caught exception: {ex_name}'
    logger.error(err)



"""A module related to calculating thermochemical properties.

Index
-----
.. currentmodule:: CAT.thermo_chem
.. autosummary::
    get_entropy
    get_thermo

API
---
.. autofunction:: get_entropy
.. autofunction:: get_thermo

"""

from typing import (Sequence, Union, Dict)

import numpy as np

from scm.plams import Molecule
from scm.plams.tools.units import Units

__all__ = ['get_thermo', 'get_entropy']

# flake8: noqa: N806,N803


def get_entropy(mol: Molecule,
                freqs: np.ndarray,
                T: float = 298.15) -> np.ndarray:
    """Calculate the translational, vibrational and rotational entropy.

    All units and constants are in SI units.

    Parameters
    ----------
    mol : |plams.Molecule|_
        A PLAMS molecule.

    freqs : |np.ndarray|_ [|np.float64|_]
        An iterable consisting of vibrational frequencies in units of cm**-1.

    T : float
        The temperature in Kelvin.

    Returns
    -------
    |np.ndarray|_ [|np.float64|_]:
        An array with translational, rotational and vibrational contributions to the entropy,
        ordered in that specific manner.
        Units are in J/mol.

    """
    # Define constants
    kT = 1.380648 * 10**-23 * T  # Boltzmann constant * temperature
    h = 6.6260701 * 10**-34  # Planck constant
    hv_kT = (h * np.asarray(freqs)) / kT  # (Planck * frequencies) / (Boltzmann * temperature)
    R = 8.31445  # Gas constant
    V_Na = ((R * T) / 10**5) / Units.constants['NA']  # Volume(1 mol ideal gas) / Avogadro's number
    pi = np.pi

    # Extract atomic masses and Cartesian coordinates
    m = np.array([at.mass for at in mol]) * 1.6605390 * 10**-27
    x, y, z = mol.as_array().T * 10**-10

    # Calculate the rotational partition function: q_rot
    inertia = np.array([
        [sum(m*(y**2 + z**2)), -sum(m*x*y), -sum(m*x*z)],
        [-sum(m*x*y), sum(m*(x**2 + z**2)), -sum(m*y*z)],
        [-sum(m*x*z), -sum(m*y*z), sum(m*(x**2 + y**2))]
    ])
    inertia_product = np.product(np.linalg.eig(inertia)[0])
    q_rot = pi**0.5 * ((8 * pi**2 * kT) / h**2)**1.5 * inertia_product**0.5

    # Calculate the translational, rotational and vibrational entropy (divided by R)
    S_trans = 1.5 + np.log(V_Na * ((2 * pi * sum(m) * kT) / h**2)**1.5)
    S_rot = 1.5 + np.log(q_rot)
    with np.errstate(divide='ignore', invalid='ignore'):
        S_vib_left = hv_kT / np.expm1(hv_kT)
        S_vib_left[np.isnan(S_vib_left)] = 0.0
        S_vib_right = np.log(1 - np.exp(-hv_kT))
        S_vib_right[S_vib_right == -np.inf] = 0.0
    S_vib = sum(S_vib_left - S_vib_right)

    return R * np.array([S_trans, S_rot, S_vib])


def get_thermo(mol: Molecule,
               freqs: Sequence[float],
               E: float = 0.0,
               T: float = 298.15,
               export: Sequence[str] = ('E', 'U', 'H', 'S', 'G'),
               unit: str = 'kcal/mol') -> Union[float, Dict[str, float]]:
    """Extract and return Gibbs free energies, entropies and/or enthalpies from an AMS KF file.

    All vibrational frequencies smaller than 100 cm**-1 are set to 100 cm**-1.

    .. _plams.Units: https://www.scm.com/doc/plams/components/utils.html#scm.plams.tools.units.Units

    Parameters
    ----------
    mol : |plams.Molecule|_
        A PLAMS molecule.

    freqs : |np.ndarray|_ [|np.float64|_]
        An iterable consisting of vibrational frequencies in units of cm**-1.

    E : float
        The eletronic energy in kcal/mol.
        Defaults to 0.0 kcal/mol.

    T : float
        The temperature in Kelvin.

    export : |tuple|_ [|str|_]
        An iterable containing strings of the to be exported energies:

        * ``'E'``: Electronic energy (see the **E** parameter)
        * ``'U'``: Interal energy (:math:`E + U_{nuc}`)
        * ``'H'``: Enthalpy (:math:`U + pV`)
        * ``'S'``: Entropy
        * ``'G'``: Gibbs free energy (:math:`H - T*S`)

    unit : str
        The unit of the to be returned energies.
        See plams.Units_ for more details and an overview of available energy units.

    Returns
    -------
    |float|_ or |dict|_ [|str|_, |float|_]:
        An energy or dictionary of energies.
        Keys

    """
    # Get frequencies; set all frequencies smaller than 100 cm**-1 to 100 cm**-1
    freqs = np.array(freqs)
    freqs[freqs < 100] = 100
    freqs *= 100 * Units.constants['c']

    # hv_kT = (Planck constant * frequencies) / (Boltzmann constant * temperature)
    hv_kT = (6.6260701 * 10**-34 * freqs) / (1.380648 * 10**-23 * T)
    RT = 8.31445 * T  # Gas constant * temperature

    # Extract and/or calculate the various energies
    E = E * Units.conversion_ratio('kcal/mol', 'kj/mol') * 1000
    U = E + RT * (3.0 + sum(0.5 * hv_kT + hv_kT / np.expm1(hv_kT)))
    H = U + RT
    S = sum(get_entropy(mol, freqs, T=T))
    G = H - T * S

    ret = {'E': E, 'U': U, 'H': H, 'S': S, 'G': G}

    if len(export) == 1:
        return Units.convert(ret[export[0]], 'kj/mol', unit) / 1000
    return {i: Units.convert(ret[i], 'kj/mol', unit) / 1000 for i in ret if i in export}



"""Functions for creating distributions of atomic indices using brute-force approaches.

Index
-----
.. currentmodule:: CAT.attachment.distribution_brute
.. autosummary::
    brute_uniform_idx

API
---
.. autofunction:: brute_uniform_idx

"""

import reprlib
from typing import Union, Callable

import numpy as np

from scm.plams import Molecule
from nanoutils import array_combinations

from .distribution import OPERATION_MAPPING

__all__ = ['brute_uniform_idx']


def brute_uniform_idx(mol: Union[Molecule, np.ndarray],
                      idx: Union[int, np.ndarray], n: int = 2,
                      operation: str = 'min',
                      weight: Callable[[np.ndarray], np.ndarray] = lambda x: np.exp(-x)
                      ) -> np.ndarray:
    r"""Brute force approach to creating uniform or clustered distributions.

    Explores, and evaluates, all valid combinations of size :math:`n` constructed
    from the :math:`k` atoms in **neighbor** closest to each atom in **center**.

    The combination where the :math:`n` atoms are closest (:code:`operation = 'max'`) or
    furthest removed from each other (:code:`operation = 'min'`) is returned.

    Parameters
    ----------
    mol : array-like [:class:`float`], shape :math:`(m,3)`
        An array-like object with Cartesian coordinate representing a collection of central atoms.

    idx : array-like [:class:`int`], shape :math:`(l,p)`
        An array-like object with indices in **mol**.
        Combinations will be explored and evaluated along axis ``-1`` of the passed array.

    n : :class:`int`
        The number of to-be returned opposing atoms.
        Should be larger than or equal to 1.

    operation : :class:`str`
        Whether to evaluate the weighted distance using :func:`argmin()<numpy.nanargmin>` or
        :func:`argmax()<numpy.nanargmax>`.
        Accepted values are ``"min"`` and ``"max"``.

    weight : :data:`Callable<typing.Callable>`
        A callable for applying weights to the distance; default: :math:`e^{-x}`.
        The callable should take an array as argument and return a new array,
        *e.g.* :func:`numpy.exp`.

    Returns
    -------
    :class:`numpy.ndarray` [:class:`int`], shape :math:`(m, n)`
        An array with indices of opposing atoms.

    See Also
    --------
    :func:`uniform_idx()<CAT.attachment.distribution.uniform_idx>`
        Yield the column-indices of **dist** which yield a uniform or clustered distribution.

    """  # noqa
    try:  # Parse **operation**
        arg_func = OPERATION_MAPPING[operation]
    except KeyError as ex:
        raise ValueError(f"Invalid value for 'operation' ({reprlib.repr(operation)}); "
                         "accepted values: ('min', 'max')") from ex

    # Find the n atoms in mol2 closest to each atom in mol1
    idx = np.array(idx, ndmin=1, copy=False)
    xyz = np.array(mol, ndmin=2, copy=False)
    if not (0 < n <= idx.shape[-1]):
        raise ValueError("'n' should be larger than 0 and smaller than or equal to the last axis of"
                         f" 'idx' ({repr(idx.shape[-1])}); observed value: {repr(n)}")

    # Evaluate all combinations of length n constructed from an iterable of size k
    idx2 = np.swapaxes(array_combinations(idx, r=n), 0, 1)
    xyz2 = xyz[idx2]

    # Construct a weighted distance matrix
    dist = np.triu(np.linalg.norm(xyz2[..., None, :] - xyz2[..., None, :, :], axis=-1))
    dist.shape = len(idx), -1, n**2
    dist_weight = weight(dist).sum(axis=-1)

    # Slice and return
    i = np.arange(len(idx))
    j = arg_func(dist_weight, axis=-1)
    return idx2[i, j]



"""A context manager for temporary interconverting between PLAMS molecules and NumPy arrays.

Index
-----
.. currentmodule:: CAT.attachment.as_array
.. autosummary::
    AsArray

API
---
.. autoclass:: AsArray
    :members:
    :private-members:
    :special-members:

"""

from typing import Iterable, Union, Sequence
from contextlib import AbstractContextManager
from collections import abc

import numpy as np

from scm.plams import Atom, Molecule

__all__ = ['AsArray']


class AsArray(AbstractContextManager):
    r"""A context manager for temporary interconverting between PLAMS molecules and NumPy arrays.

    Examples
    --------
    .. code:: python

        >>> from scm.plams import Molecule

        # Create a H2 example molecule
        >>> h1 = Atom(symbol='H', coords=(0.0, 0.0, 0.0))
        >>> h2 = Atom(symbol='H', coords=(1.0, 0.0, 0.0))
        >>> mol = Molecule()
        >>> mol.add_atom(h1)
        >>> mol.add_atom(h2)

        >>> print(mol)  # doctest: +SKIP
          Atoms:
            1         H      0.000000      0.000000      0.000000
            2         H      1.000000      0.000000      0.000000
        <BLANKLINE>

        # Example: Translate the molecule along the Cartesian Z-axis by 5 Angstroem
        >>> with AsArray(mol) as xyz:
        ...     xyz[:, 2] += 5

        >>> print(mol)  # doctest: +SKIP
          Atoms:
            1         H      0.000000      0.000000      5.000000
            2         H      1.000000      0.000000      5.000000
        <BLANKLINE>

    Parameters
    ----------
    mol : |plams.Molecule| or |Iterable| [|plams.Atom|]
        An iterable consisting of PLAMS atoms.
        See :attr:`AsArray.mol`.

    Attributes
    ----------
    mol : |plams.Molecule| or |Sequence| [|plams.Atom|]
        A PLAMS molecule or a sequence of PLAMS atoms.

    _xyz : :math:`n*3` :class:`numpy.ndarray` [:class:`float`], optional
        A 2D array with the Cartesian coordinates of **mol**.
        Empty by default; this value is set internally by the :meth:`AsArray.__enter__` method.

    """

    @property
    def mol(self) -> Union[Molecule, Sequence[Atom]]:
        """Get or set the embedded molecule."""
        return self._mol

    @mol.setter
    def mol(self, value: Iterable[Atom]) -> None:
        self._mol = value if isinstance(value, (abc.Sequence, Molecule)) else tuple(value)

    def __init__(self, mol: Iterable[Atom]) -> None:
        """Initialize a :class:`AsArray` instance."""
        self.mol = mol
        self._xyz = None

    def __enter__(self) -> np.ndarray:
        """Enter the context manager; return an array of Cartesian coordinates."""
        self._xyz = Molecule.as_array(None, atom_subset=self.mol)
        return self._xyz

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context manager; update the Cartesian coordinates of :attr:`AsArray.mol`."""
        Molecule.from_array(None, self._xyz, atom_subset=self.mol)
        self._xyz = None



"""A module designed for sanitizing and interpreting the input file.

Index
-----
.. currentmodule:: CAT.data_handling.validate_input
.. autosummary::
    validate_input

API
---
.. autofunction:: validate_input

"""

import sys
from os import mkdir
from os.path import (join, isdir)
from typing import Dict, Any

if sys.version_info >= (3, 7):
    from contextlib import nullcontext
else:
    from contextlib2 import nullcontext

import qmflows
from scm.plams import Settings, Cp2kJob

from .validation_schemas import (
    core_schema,
    ligand_schema,
    qd_schema,
    database_schema,
    mongodb_schema,
    bde_schema,
    qd_opt_schema,
    crs_schema,
    asa_schema,
    ligand_opt_schema,
    subset_schema,
    multi_ligand_schema,
    cdft_schema,
    bulkiness_schema,
)

from .validate_ff import validate_ff, update_ff_jobs
from .validate_mol import validate_mol
from .anchor_parsing import parse_anchors
from ..utils import validate_path, SetEnviron
from ..logger import logger

try:
    from dataCAT import Database
    DATA_CAT = True
except ImportError:
    DATA_CAT = False

__all__ = ['validate_input']


def _validate_multi_lig(s: Settings) -> None:
    """Check that one (and only one!) of ``'f'`` and ``'dummy'`` is specified."""
    f = s.optional.qd.multi_ligand.f

    anchor = s.optional.qd.multi_ligand.anchor
    if anchor is None:
        anchor = s.optional.qd.multi_ligand.dummy
        s.optional.qd.multi_ligand.anchor = anchor
    del s.optional.qd.multi_ligand.dummy

    if f is anchor is None:
        raise ValueError("'.multi_ligand.f' and '.multi_ligand.anchor' cannot be "
                         "both unspecified or set to 'None'")
    elif None not in (f, anchor):
        raise ValueError("Only one of '.multi_ligand.f' and '.multi_ligand.anchor' "
                         "should be specified")

    if anchor is not None:
        assert len(anchor) == len(s.optional.qd.multi_ligand.ligands)
    else:
        assert len(f) == len(s.optional.qd.multi_ligand.ligands) - 1


def parse_qmflows_keywords(bde_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Parse qmflows generic keywords for the BDE workflow."""
    # TODO: Support QM packages other than qmflows
    if bde_dict["job1"] is Cp2kJob:
        s1 = qmflows.Settings(bde_dict["s1"])
        s1 = qmflows.cp2k.generic2specific(s1)
        if "specific" in s1:
            s1.input.update(s1.pop("specific").pop(qmflows.cp2k.pkg_name))
        bde_dict["s1"] = s1

    if bde_dict["job2"] is Cp2kJob:
        s2 = qmflows.Settings(bde_dict["s2"])
        s2 = qmflows.cp2k.generic2specific(s2)
        if "specific" in s1:
            s2.input.update(s2.pop("specific").pop(qmflows.cp2k.pkg_name))
        bde_dict["s2"] = s2
    return bde_dict


def validate_input(s: Settings, validate_only: bool = True) -> None:
    """Initialize the input-validation procedure.

    performs an inplace update of **s**.

    Parameters
    ----------
    s : |plams.Settings|_
        A Settings instance with to-be validated CAT input settings.
    validate_only : bool
        Perform only validation.

    """
    dirnames = ('database', 'ligand', 'core', 'qd')
    if not validate_only:
        # Validate the path
        s.path = validate_path(s.get("path"))

        # Set the various working directories
        for key in dirnames:
            value = join(s.path, key)
            s.optional[key].dirname = value
            if not isdir(value):
                mkdir(value)
    else:
        if s.path is None:
            s.path = '.'
        for key in dirnames:
            s.optional[key].dirname = join(s.path, key)

    # Validate optional argument
    if not validate_only:
        context = nullcontext()
    else:
        context = SetEnviron(AMSBIN='', AMSHOME='', AMSRESOURCES='', SCMLICENSE='')

    s.optional.database = database_schema.validate(s.optional.database)
    with context:
        s.optional.ligand = ligand_schema.validate(s.optional.ligand)
        s.optional.core = core_schema.validate(s.optional.core)
        s.optional.qd = qd_schema.validate(s.optional.qd)

        # Validate some of the more complex optionala rguments
        if s.optional.database.mongodb:
            s.optional.database.mongodb = mongodb_schema.validate(s.optional.database.mongodb)

        if s.optional.core.subset:
            s.optional.core.subset = subset_schema.validate(s.optional.core.subset)
            if 'p' in s.optional.core.subset:
                if 'weight' in s.optional.core.subset:
                    raise KeyError("'p' and 'weight' cannot be simultaneously specified")
                logger.warn("The 'subset.p' parameter is deprecated; see 'subset.weight'")
                p = s.optional.core.subset.pop('p')
                s.optional.core.subset.weight = lambda x: -(x**p)
        if s.optional.core.anchor is not None:
            pass
        elif s.optional.core.dummy is not None:
            s.optional.core.anchor = s.optional.core.dummy
        else:
            s.optional.core.anchor = 17
        del s.optional.core.dummy

        if s.optional.ligand.optimize:
            s.optional.ligand.optimize = ligand_opt_schema.validate(s.optional.ligand.optimize)
        if s.optional.ligand.cdft:
            s.optional.ligand.cdft = cdft_schema.validate(s.optional.ligand.cdft)

        if s.optional.get('crs') is not None:
            s.optional.ligand.crs = crs_schema.validate(s.optional.crs)
        elif s.optional.ligand['cosmo-rs']:
            crs = s.optional.ligand.pop('cosmo-rs')
            s.optional.ligand.crs = crs_schema.validate(crs)

        if s.optional.qd.optimize:
            s.optional.qd.optimize = qd_opt_schema.validate(s.optional.qd.optimize)
        if s.optional.qd.dissociate:
            bde_dict = bde_schema.validate(s.optional.qd.dissociate)
            if (
                (bde_dict["core_atom"] is bde_dict["lig_count"] is None) or
                (bde_dict["core_atom"] is not None and bde_dict["lig_count"] is not None)
            ):
                raise ValueError("One of `core_atom` and `lig_count` must be specified")
            s.optional.qd.dissociate = parse_qmflows_keywords(bde_dict)
        if s.optional.qd.activation_strain:
            s.optional.qd.activation_strain = asa_schema.validate(s.optional.qd.activation_strain)
        if s.optional.qd.multi_ligand:
            s.optional.qd.multi_ligand = multi_ligand_schema.validate(s.optional.qd.multi_ligand)
            _validate_multi_lig(s)
        if s.optional.qd.bulkiness:
            s.optional.qd.bulkiness = bulkiness_schema.validate(s.optional.qd.bulkiness)

        # Create forcefield Job Settings
        if s.optional.forcefield:
            s.optional.forcefield = validate_ff(s.optional.forcefield)
            update_ff_jobs(s)

        # Validate the input cores and ligands
        validate_mol(s.input_cores, 'input_cores', join(s.path, 'core'), not validate_only)
        validate_mol(s.input_ligands, 'input_ligands', join(s.path, 'ligand'), not validate_only)
        if 'input_qd' in s:
            validate_mol(s.input_qd, 'input_qd', join(s.path, 'qd'), not validate_only)

        # Create a dataCAT.Database instance
        if s.optional.database.get('db') is None and not validate_only:
            if DATA_CAT:
                db_path = s.optional.database.dirname
                s.optional.database.db = Database(path=db_path, **s.optional.database.mongodb)
            else:
                s.optional.database.db = None

        # Create RDKit molecules representing functional groups
        if s.optional.ligand.anchor is not None:
            func_groups = s.optional.ligand.anchor
        else:
            func_groups = s.optional.ligand.functional_groups
        del s.optional.ligand.functional_groups

        split = s.optional.ligand.split
        s.optional.ligand.anchor = parse_anchors(func_groups, split=split)
        s.optional.core.anchor = parse_anchors(s.optional.core.anchor, split=True, is_core=True)



"""A module for storing all column keys used throughout CAT.

Index
-----
.. currentmodule:: CAT.workflows.key_map
.. autosummary::
    KEY_MAP

API
---
.. autodata:: KEY_MAP
    :annotation: : Mapping[str, Tuple[str, str]]

"""

from types import MappingProxyType
from typing import Mapping, Tuple

#: A :class:`Mapping<collections.abc.Mapping>` containing all column keys used throughout CAT.
KEY_MAP: Mapping[str, Tuple[str, str]] = MappingProxyType({
    'OPT': ('opt', ''),
    'MOL': ('mol', ''),
    'FORMULA': ('formula', ''),
    'HDF5_INDEX': ('hdf5 index', ''),
    'JOB_SETTINGS_QD_OPT': ('job_settings_qd_opt', ''),
    'JOB_SETTINGS_CRS': ('job_settings_crs', ''),
    'JOB_SETTINGS_BDE': ('job_settings_BDE', ''),
    'JOB_SETTINGS_ASA': ('job_settings_ASA', ''),
    'JOB_SETTINGS_CDFT': ('job_settings_cdft', ''),
    'CDFT_MU': ('cdft', 'Electronic chemical potential (mu)'),
    'CDFT_CHI': ('cdft', 'Electronegativity (chi=-mu)'),
    'CDFT_ETA': ('cdft', 'Hardness (eta)'),
    'CDFT_S': ('cdft', 'Softness (S)'),
    'CDFT_GAMMA': ('cdft', 'Hyperhardness (gamma)'),
    'CDFT_OMEGA': ('cdft', 'Electrophilicity index (w=omega)'),
    'CDFT_NUCLEOFUGE': ('cdft', 'Dissocation energy (nucleofuge)'),
    'CDFT_ELECTROFUGE': ('cdft', 'Dissociation energy (electrofuge)'),
    'CDFT_W_MINUS': ('cdft', 'Electrodonating power (w-)'),
    'CDFT_W_PLUS': ('cdft', 'Electroaccepting power(w+)'),
    'CDFT_ELECTROPHILICITY': ('cdft', 'Net Electrophilicity'),
    'CDFT_DELTAF_MINUS': ('cdft', 'Global Dual Descriptor Deltaf-'),
    'CDFT_DELTAF_PLUS': ('cdft', 'Global Dual Descriptor Deltaf+'),
    'CDFT_MU_MINUS': ('cdft', 'Electronic chemical potential (mu-)'),
    'CDFT_MU_PLUS': ('cdft', 'Electronic chemical potential (mu+)'),
    'ASA_INT': ('ASA', 'E_int'),
    'ASA_STRAIN': ('ASA', 'E_strain'),
    'ASA_E': ('ASA', 'E'),
    'SETTINGS1': ('settings', '1'),
    'SETTINGS2': ('settings', '2'),
    'SETTINGS_SOLV1': ('settings', 'solv 1'),
    'SETTINGS_SOLV2': ('settings', 'solv 2'),
    'SETTINGS_ASA': ('settings', 'ASA 1'),
    'SETTINGS_BDE1': ('settings', 'BDE 1'),
    'SETTINGS_BDE2': ('settings', 'BDE 2'),
    'SETTINGS_CDFT': ('settings', 'cdft 1'),
    'V_BULK': ('V_bulk', '')
})

globals().update(KEY_MAP)

__all__ = list(KEY_MAP.keys())



"""A module designed for sanitizing and validating the ``["optional"]["forcefield"]`` input block.

Index
-----
.. currentmodule:: CAT.data_handling.validate_ff
.. autosummary::
    validate_ff

API
---
.. autofunction:: validate_ff

"""  # noqa: E501

import reprlib
from typing import Union, Any, Mapping, TypeVar, Optional
from collections import abc

from scm.plams import Settings, Cp2kJob

from CAT.utils import get_template

try:
    from nanoCAT.ff.cp2k_utils import set_cp2k_param
except ImportError as ex:
    NANO_CAT: Optional[ImportError] = ex
else:
    NANO_CAT: Optional[ImportError] = None

__all__ = ['validate_ff', 'update_ff_jobs']

M = TypeVar('M', bound=Mapping)


def validate_ff(s: M) -> Union[M, bool]:
    """Validate the ``["optional"]["forcefield"]`` block in the CAT input.

    Return ``False`` if an empty dictionary is provided.

    Example
    -------
    An example of a valid input block:

    .. code::

        optional:
            forcefield:
                charge:
                    keys: [input, force_eval, mm, forcefield, charge]
                    Cd: 0.9768
                    Se: -0.9768
                epsilon:
                    unit: kjmol
                    keys: [input, force_eval, mm, forcefield, nonbonded, lennard-jones]
                    Cd Cd: 0.3101
                    Se Se: 0.4266
                    Cd Se: 1.5225
                sigma:
                    unit: nm
                    keys: [input, force_eval, mm, forcefield, nonbonded, lennard-jones]
                    Cd Cd: 0.1234
                    Se Se: 0.4852
                    Cd Se: 0.2940

    """
    if not s:
        return False

    for k1, v1 in s.items():
        assert isinstance(k1, str), assertion_msg(k1)
        assert isinstance(v1, dict), assertion_msg(v1)

        assert 'keys' in v1
        assert isinstance(v1['keys'], abc.Sequence), assertion_msg(v1['keys'])
        v1['keys'] = tuple(v1['keys'])

        if 'unit' in v1:
            assert isinstance(v1['unit'], str), assertion_msg(v1['unit'])

        for k2, v2 in v1.items():
            if k2 in ('unit', 'keys'):
                continue
            assert isinstance(k2, str), assertion_msg(k2)
            assert isinstance(v2, (str, int, float)), assertion_msg(v2)

    return s


def assertion_msg(obj: Any) -> str:
    """Return a formatted string."""
    return f'<type {repr(obj.__class__.__name__)}: {reprlib.repr(obj)}>'


def update_ff_jobs(s: Settings) -> None:
    """Update forcefield settings."""
    if NANO_CAT is not None:
        raise NANO_CAT

    ff = Settings()
    set_cp2k_param(ff, s.optional.forcefield)

    optimize = s.optional.qd.optimize
    if optimize and optimize.use_ff:
        if optimize.job1 and str(optimize.job1) == str(Cp2kJob):
            optimize.s1 = Settings() if optimize.s1 is None else optimize.s1
            optimize.s1 += get_template('qd.yaml')['CP2K_CHARM_opt']
            optimize.s1 += ff

        if optimize.job2 and str(optimize.job1) == str(Cp2kJob):
            optimize.s2 = Settings() if optimize.s2 is None else optimize.s2
            optimize.s2 += get_template('qd.yaml')['CP2K_CHARM_opt']
            optimize.s2 += ff

    dissociate = s.optional.qd.dissociate
    if dissociate and dissociate.use_ff:
        if dissociate.job1 and str(dissociate.job1) == str(Cp2kJob):
            dissociate.s1 = Settings() if dissociate.s1 is None else dissociate.s1
            dissociate.s1 += get_template('qd.yaml')['CP2K_CHARM_opt']
            dissociate.s1 += ff

    activation_strain = s.optional.qd.activation_strain
    if activation_strain and activation_strain.use_ff:
        if activation_strain.job1 and str(activation_strain.job1) == str(Cp2kJob):
            key = 'CP2K_CHARM_singlepoint' if not activation_strain.md else 'CP2K_CHARM_md'
            activation_strain.s1 = Settings() if activation_strain.s1 is None else activation_strain.s1  # noqa
            activation_strain.s1 += get_template('qd.yaml')[key]
            activation_strain.s1 += ff



"""A module for holding the :class:`.SettingsDataFrame` and :class:`.SettingsSeries` classes.

Index
-----
.. currentmodule:: CAT.settings_dataframe
.. autosummary::
    SettingsSeries
    SettingsDataFrame

API
---
.. autoclass:: CAT.settings_dataframe.SettingsSeries
    :members:
    :private-members:
    :special-members:

.. autoclass:: CAT.settings_dataframe.SettingsDataFrame
    :members:
    :private-members:
    :special-members:

"""

from typing import Optional, Callable

import pandas as pd

from .frozen_settings import FrozenSettings

__all__ = ['SettingsSeries', 'SettingsDataFrame']


class SettingsSeries(pd.Series):
    """A subclass of the Pandas Series with an additional :attr:`.settings` attribute.

    Parameters
    ----------
    settings : dict
        Optional: A dictionary with additional user-defined (meta-)data.
        See :attr:`.SettingsSeries.settings`.

    Attributes
    ----------
    settings : |CAT.FrozenSettings|_
        An immutable :class:`.FrozenSettings` instance with additional user-defined (meta-)data.

    """

    _metadata = ['settings']

    def __init__(self, data=None, index=None, dtype=None, name=None, copy=False, fastpath=False,
                 settings: Optional[dict] = None) -> None:
        """Initialize the :class:`.SettingsSeries` construction."""
        self.settings = self._sanitize_settings(settings)
        super().__init__(data, index, dtype, name, copy, fastpath)

    @property
    def _constructor(self) -> Callable[..., 'SettingsSeries']:
        """Construct a :class:`.SettingsSeries` instance."""
        def _series(*args, **kwargs) -> SettingsSeries:
            return SettingsSeries(*args, **kwargs).__finalize__(self)
        return _series

    @property
    def _constructor_expanddim(self) -> Callable[..., 'SettingsDataFrame']:
        """Construct a :class:`.SettingsDataFrame` instance."""
        def _df(*args, **kwargs) -> SettingsDataFrame:
            return SettingsDataFrame(*args, **kwargs).__finalize__(self)
        return _df

    @staticmethod
    def _sanitize_settings(settings: Optional[dict]) -> FrozenSettings:
        """Sanitize the **settings** parameter for :attr:`SettingsSeries.settings`."""
        if settings is None:
            return FrozenSettings()
        elif isinstance(settings, dict):
            return FrozenSettings(settings)
        else:
            err = "The settings argument expects an instance of 'dict'; observed type '{}'"
            raise TypeError(err.format(settings.__class__.__name__))


class SettingsDataFrame(pd.DataFrame):
    """A subclass of the Pandas DataFrame with an additional :attr:`.settings` attribute.

    Parameters
    ----------
    settings : dict
        Optional: A dictionary with additional user-defined (meta-)data.
        See :attr:`.SettingsDataFrame.settings`.

    Attributes
    ----------
    settings : |CAT.FrozenSettings|_
        An immutable :class:`.FrozenSettings` instance with additional user-defined (meta-)data.

    """

    _metadata = ['settings']

    def __init__(self, data=None, index=None, columns=None, dtype=None, copy=False,
                 settings: Optional[dict] = None) -> None:
        """Initialize the :class:`.SettingsDataFrame` construction."""
        self.settings = self._sanitize_settings(settings)
        super().__init__(data, index, columns, dtype, copy)

    @property
    def _constructor(self) -> Callable[..., 'SettingsDataFrame']:
        """Construct a :class:`.SettingsDataFrame` instance."""
        def _df(*args, **kwargs) -> SettingsDataFrame:
            return SettingsDataFrame(*args, **kwargs).__finalize__(self)
        return _df

    @property
    def _constructor_sliced(self) -> Callable[..., 'SettingsSeries']:
        """Construct a :class:`.SettingsSeries` instance."""
        def _series(*args, **kwargs) -> SettingsSeries:
            return SettingsSeries(*args, **kwargs).__finalize__(self)
        return _series

    @staticmethod
    def _sanitize_settings(settings: Optional[dict]) -> FrozenSettings:
        """Sanitize the **settings** parameter for :attr:`SettingsDataFrame.settings`."""
        if settings is None:
            return FrozenSettings()
        elif isinstance(settings, dict):
            return FrozenSettings(settings)
        else:
            err = "The settings argument expects an instance of 'dict'; observed type '{}'"
            raise TypeError(err.format(settings.__class__.__name__))



"""A module designed for exporting PLAMS :class:`Molecule` instances to common file formats.

Index
-----
.. currentmodule:: CAT.data_handling.mol_to_file
.. autosummary::
    mol_to_file

API
---
.. autofunction:: mol_to_file

"""

import os
from types import MappingProxyType
from typing import Iterable, Container, Union, Callable
from os.path import join, isdir, isfile, exists

from scm.plams import Molecule, writepdb

from ..logger import logger

__all__ = ['mol_to_file']

MolExportFunc = Callable[[Molecule, Union[str, bytes, os.PathLike]], None]

#: A mapping of file extensions to a Callable for Molecule exporting
EXPORT_MAPPING: "MappingProxyType[str, MolExportFunc]" = MappingProxyType({
    'pdb': writepdb,
    'xyz': Molecule.write,
    'mol': Molecule.write,
    'mol2': Molecule.write
})


def mol_to_file(
    mol_list: Iterable[Molecule],
    path: "None | str" = None,
    overwrite: bool = True,
    mol_format: Container[str] = ('xyz', 'pdb'),
) -> None:
    """Export all molecules in **mol_list** to .pdb, .xyz, .mol or .mol2 files.

    Parameters
    ----------
    mol_list: |list|_ [|plams.Molecule|_]
        An iterable consisting of PLAMS molecules.

    path : str
        Optional: The path to the directory where the molecules will be stored.
        Defaults to the current working directory if ``None``.

    overwrite : bool
        If previously generated files can be overwritten or not.

    mol_format : |list|_ [|str|_]
        A list of strings with the to-be exported file types.
        Accepted values are ``"xyz"``, ``"pdb"``, ``"mol"`` and/or ``"mol2"``.

    Raises
    ------
    FileNotFoundError
        Raised **path** does not exist.

    NotADirectoryError
        Raised **path** is not a directory.

    """
    # Set the export path
    _path = path or os.getcwd()
    if not exists(_path):
        raise FileNotFoundError(f'{path} not found')
    elif not isdir(_path):
        raise NotADirectoryError(f'{path} is not a directory')

    condition = isfile if overwrite else lambda n: True
    export_dict = {k: v for k, v in EXPORT_MAPPING.items() if k in mol_format}
    if not export_dict:
        raise ValueError("No valid values found in the mol_format argument; accepted values are: "
                         "'xyz', 'pdb', 'mol' and/or 'mol2'")

    for mol in mol_list:
        mol_path = join(_path, mol.properties.name)
        if not condition:
            continue

        try:
            for ext, func in export_dict.items():
                filename = f'{mol_path}.{ext}'
                func(mol, filename)
        except OSError as ex:
            if ex.errno == 36:
                if logger is None:
                    continue
                logger.error(f"Filename too long: failed to create {filename!r}", exc_info=True)
            else:
                raise



"""Functions for creating distance matrices.

Index
-----
.. currentmodule:: CAT.attachment.edge_distance
.. autosummary::
    edge_dist
    array_combinations
    to_convex

API
---
.. autofunction:: edge_dist
.. autofunction:: array_combinations
.. autofunction:: to_convex

"""

import reprlib
from typing import Optional, Any

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import dijkstra
from scipy.spatial import ConvexHull

from nanoutils import array_combinations

try:
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

    PLT: Optional[ImportError] = None
    Figure = plt.Figure
except ImportError as ex:
    PLT = ex
    Figure = 'matplotlib.pyplot.Figure'

__all__ = ['edge_dist', 'plot_polyhedron']


def to_convex(xyz: np.ndarray, n: float = 1.0) -> np.ndarray:
    r"""Round all edges in **xyz** by a factor **n**: (:math:`0 < n \le 1`)."""
    if not (0.0 < n <= 1.0):
        raise ValueError(f"Expected '0 < n <= 1'; observed: {reprlib.repr(n)}")
    xyz = np.asarray(xyz, dtype=float)
    xyz = xyz - xyz.mean(axis=0)

    r = np.linalg.norm(xyz, axis=-1)
    dr = n * (r.mean() - r)

    elongate = (dr + r) / r
    xyz *= elongate[..., None]
    return xyz


def edge_dist(xyz: np.ndarray, n: float = 1.0,
              edges: Optional[np.ndarray] = None) -> np.ndarray:
    r"""Calculate all shortest paths between all points in the polyhedron **xyz** by traversing its edges.

    After converting **xyz** into a polyhedron with triangular faces,
    using a convex hull algorithm,
    the shortest paths between all possible point pairs is calculated using
    the Dijkstra algorithm.
    Paths are constrained to the edges of the polyhedron,
    as opposed to a "normal" distance matrix each element thus represents
    the shortest path along a set of 1D lines rather than a 3D volume.

    Given the matrix of Cartesian coordinates :math:`\boldsymbol{X} \in \mathbb{R}^{n, 3}`,
    the matching distance matrix :math:`\boldsymbol{D} \in \mathbb{R}^{n, n}` and
    the matching edge-distance matrix :math:`\boldsymbol{D}^{\text{edge}} \in \mathbb{R}^{n, n}`,
    then element :math:`D_{i,j}` is defined as following:

    .. math::

        D_{i, j} = ||\boldsymbol{X}_{i,:} - \boldsymbol{X}_{j,:}||

    while :math:`D_{i, j}^{\text{edge}}` is defined as:

    .. math::

        D_{i, j}^{\text{edge}} = \min_{\boldsymbol{p} \in \mathbb{N}^{m}; m \in \mathbb{N}}
        \sum_{k=0}^{m-1} || X_{p_{k},:} - X_{p_{k+1},:} ||
        \quad \text{with} \quad p_{0} = i \quad \text{and} \quad p_{m} = j

    The vector :math:`\boldsymbol{p}` is a path,
    represented by the indices of neighbouring vertices in :math:`\boldsymbol{X}`,
    whose length (*i.e.* sum of Euclidean distances) is to-be minimized.

    Notes
    -----
    All points in **xyz** are projected on the surface of a sphere during the construction
    of the convex hull (if :code:`n != 0.0`);
    the quality of the constructed polyhedron will thus depend on the "convexness" of **xyz**.
    For highly concave structures (*e.g.* a torus) it is strongly recommended to
    manually pass all edge indices using the **edges** parameter.

    Parameters
    ----------
    xyz : array-like [:class:`float`], shape :math:`(m, 3)`
        A 2D array-like object of Cartesian coordinates representing a polyhedron.
        The supplied polyhedron should be convex in shape.

    n : :class:`float`
        Smoothing factor for constructing a convex hull.
        Should obey :math:`0 <= n <= 1`.

    edges : array-like [:class:`int`], shape :math:`(n, 2)`, optional
        A 2D array-like object with all indice-pairs in **xyz** representing polyhedron edges.

    Returns
    -------
    :math:`(m, m)` :class:`numpy.ndarray`
        A 2D array containing all possible (Euclidean) distance-pairs in **xyz**.
        Distances are calculated by traversing the shortest path along the edges of **xyz**,
        rather than moving directly through space.

    See Also
    --------
    :class:`ConvexHull<scipy.spatial.ConvexHull>`
        Convex hulls in N dimensions.

    :func:`dijkstra<scipy.sparse.csgraph.dijkstra>`
        Dijkstra algorithm using Fibonacci Heaps.

    :func:`cdist<scipy.spatial.distance.cdist>`
        Compute distance between each pair of the two collections of inputs.

    """  # noqa
    xyz = np.array(xyz, dtype=float, ndmin=2, copy=False)
    xyz_len = len(xyz)

    # Create an array with all index-pairs forming the polyhedron edges
    if edges is None:
        xyz_convex = to_convex(xyz, n=n) if not np.allclose(n, 0.0) else xyz
        hull = ConvexHull(xyz_convex)
        idx = array_combinations(hull.simplices, r=2)
        idx.shape = -1, 2
    else:
        idx = np.asarray(edges, dtype=int)
    i, j = idx.T

    # Create a distancea matrix containing all edge distances
    dist = np.zeros((xyz_len, xyz_len), dtype=float)
    dist[i, j] = dist[j, i] = np.linalg.norm(xyz[i] - xyz[j], axis=1)
    dist_sparse = csr_matrix(dist)

    # Traverse the edges
    return dijkstra(dist_sparse,
                    directed=False,
                    indices=np.arange(xyz_len),
                    return_predecessors=False)


def plot_polyhedron(xyz: np.ndarray, triangles: Optional[np.ndarray] = None,
                    show: bool = True, **kwargs: Any) -> Figure:
    r"""Plot a polyhedron, represented by an array of Cartesian coordinates, with matplotlib.

    Parameters
    ----------
    xyz : array-like [:class:`float`], shape :math:`(m, 3)`
        A 2D array-like object representing the Cartesian coordinates of a polyhedron.

    triangles : array-like [:class:`int`], shape :math:`(n, 3)`, optional
        A 2D array-like object with all indice-pairs in **xyz** representing the triangular
        faces of the **xyz** polygon.

    show : :class:`bool`
        Show the created figure.

    \**kwargs : :data:`Any<typing.Any>`
        Further keyword arguments for
        :meth:`Axes.plot_trisurf()<matplotlib.pyplot.Axes.plot_trisurf>`.

    Returns
    -------
    :class:`Figure<matplotlib.pyplot.Figure>`
        The resulting matplotlib Figure.

    """
    if PLT is not None:
        raise PLT

    if 'cmap' not in kwargs:
        kwargs['cmap'] = plt.cm.Spectral

    xyz = np.asarray(xyz, dtype=float)
    triangles = ConvexHull(xyz).simplices if triangles is None else np.asarray(triangles, dtype=int)

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    ax.plot_trisurf(*xyz.T, triangles=triangles, **kwargs)

    if show:
        plt.show(block=True)
    return fig



"""A module for attaching multiple non-unique ligands to a single quantum dot."""

import os
from typing import Iterable, Any, overload, Sequence, MutableSequence, List, Union

import numpy as np
import pandas as pd

from scm.plams import Molecule, MoleculeError

from .utils import AnchorTup, get_formula
from .workflows import WorkFlow
from .mol_utils import to_symbol
from .data_handling import mol_to_file
from .data_handling.mol_import import read_mol
from .data_handling.validate_mol import validate_mol
from .attachment.ligand_opt import optimize_ligand, allign_axis
from .attachment.ligand_anchoring import find_substructure
from .attachment.ligand_attach import ligand_to_qd

__all__ = ['init_multi_ligand']


def init_multi_ligand(qd_df):
    """Initialize the multi-ligand attachment procedure."""
    workflow = WorkFlow.from_template(qd_df, name='multi_ligand')

    if workflow.anchor is not None:
        sequence = [to_symbol(i) for i in workflow.anchor]
    elif workflow.f is not None:
        sequence = [str(i) for i in workflow.f]
    else:
        raise TypeError("'workflow.f' and 'workflow.anchor' cannot be both 'None'")

    columns_iter1 = ('/'.join(item for item in sequence[:i]) for i in range(1, 1+len(sequence)))
    columns_iter2 = (('multi ligand', i) for i in columns_iter1)
    columns = pd.MultiIndex.from_tuples(columns_iter2, names=qd_df.columns.names)
    for k in columns:
        qd_df[k] = None

    workflow(multi_lig, qd_df, columns=columns, no_loc=True)

    if workflow.mol_format is not None:
        path = workflow.path
        mol_format = workflow.mol_format
        mol_ar_flat = qd_df[columns].values.ravel()
        mol_to_file(mol_ar_flat, path, mol_format=mol_format)


@overload
def multi_lig(qd_series: pd.Series, ligands: Iterable[str],
              anchor: Sequence[Union[str, int]], f: None,
              **kwargs: Any) -> pd.DataFrame:
    ...
@overload  # noqa: E302
def multi_lig(qd_series: pd.Series, ligands: Iterable[str],
              anchor: None, f: Sequence[float],
              **kwargs: Any) -> pd.DataFrame:
    ...
def multi_lig(qd_series, ligands, anchor=None, f=None, **kwargs):  # noqa: E302
    """Attach multiple non-unique **ligands** to each qd in **qd_series**."""
    # Read and parse the SMILES strings
    ligands = smiles_to_lig(list(ligands),
                            functional_groups=kwargs['functional_groups'],
                            opt=kwargs['opt'],
                            split=kwargs['split'])

    mol_format = kwargs.get('mol_format')
    _path = kwargs.get('path')
    if mol_format is not _path is not None:
        path = os.path.join(str(_path).rsplit(os.sep, 1)[0], 'ligand')
        mol_to_file(ligands, path, mol_format=mol_format)

    if f is not None:
        raise NotImplementedError("'f != None' is not yet implemented")

    if anchor is not None:
        return _multi_lig_anchor(qd_series, ligands, kwargs['path'], anchor, kwargs['allignment'])
    elif f is not None:
        return [[NotImplemented]]
    else:
        raise TypeError("'f' and 'anchor' cannot be both 'None'")


def _multi_lig_anchor(qd_series, ligands, path, anchor, allignment) -> np.ndarray:
    """Gogogo."""
    ret = np.empty((len(ligands), len(qd_series)), dtype=object)
    for i, qd in enumerate(qd_series):
        qd = qd.copy()

        for j, (ligand, atnum) in enumerate(zip(ligands, anchor)):
            qd.set_atoms_id()
            try:
                atoms = [at for at in qd if at.atnum == atnum]
                assert atoms
            except AssertionError as ex:
                raise MoleculeError(f'Failed to identify {to_symbol(atnum)!r} in '
                                    f'{get_formula(qd)!r}') from ex

            coords = Molecule.as_array(None, atom_subset=atoms)
            qd.properties.dummies = np.array(coords, ndmin=2, dtype=float)
            qd = ligand_to_qd(qd, ligand, path=path,
                              allignment=allignment,
                              idx_subset=qd.properties.indices)
            ret[j, i] = qd
            for at in reversed(atoms):
                qd.delete_atom(qd[at.id])
            qd.unset_atoms_id()
    return ret


def _multi_lig_f(qd_series, ligands, path, f, **kwargs):
    raise NotImplementedError


def smiles_to_lig(smiles: MutableSequence[str],
                  functional_groups: Iterable[AnchorTup],
                  opt: bool = True, split: bool = True) -> List[Molecule]:
    """Parse and convert all **smiles** strings into Molecules."""
    # Convert the SMILES strings into ligands
    validate_mol(smiles, 'input_ligands')
    ligands = [find_substructure(lig, functional_groups, split)[0] for lig in read_mol(smiles)]

    # Optimize the ligands
    process_mol = optimize_ligand if opt else lambda n: allign_axis(n)
    for lig in ligands:
        process_mol(lig)
    return ligands



"""**CAT**: A collection of tools designed for the construction of various chemical compounds."""  # noqa: E501

# flake8: noqa: E402

from .__version__ import __version__ as __version__
__author__ = "Bas van Beek"
__email__ = 'b.f.van.beek@vu.nl'

from nanoutils import VersionInfo
version_info = VersionInfo.from_str(__version__)
del VersionInfo

from .logger import logger
from .jobs import (
    job_single_point, job_geometry_opt, job_freq
)
from .thermo_chem import (
    get_thermo, get_entropy
)
from .attachment import (
    init_qd_opt, init_ligand_opt, init_qd_construction, init_ligand_anchoring
)
from .data_handling import (
    read_mol, set_mol_prop
)
from .utils import get_template

from . import distribution

__all__ = [
    'logger',

    'job_single_point', 'job_geometry_opt', 'job_freq',

    'get_thermo', 'get_entropy',

    'init_qd_opt', 'init_ligand_opt', 'init_qd_construction', 'init_ligand_anchoring',

    'read_mol', 'set_mol_prop',

    'get_template',
]



"""A module handling the interaction with all other modules, functioning as recipe.

Index
-----
.. currentmodule:: CAT.base
.. autosummary::
    prep
    prep_input
    prep_core
    prep_ligand
    prep_qd
    val_nano_cat

API
---
.. autofunction:: prep
.. autofunction:: prep_input
.. autofunction:: prep_core
.. autofunction:: prep_ligand
.. autofunction:: prep_qd
.. autofunction:: val_nano_cat

"""

from time import time
from typing import Optional, Tuple

import numpy as np
import pandas as pd

from scm.plams import Settings, MoleculeError

from .__version__ import __version__

from .logger import logger
from .settings_dataframe import SettingsDataFrame

from .data_handling.mol_import import read_mol
from .data_handling.update_qd_df import update_qd_df
from .data_handling.validate_input import validate_input

from .multi_ligand import init_multi_ligand
from .attachment.qd_opt import init_qd_opt
from .attachment.ligand_opt import init_ligand_opt, allign_axis
from .attachment.ligand_attach import init_qd_construction
from .attachment.ligand_anchoring import init_ligand_anchoring
from .attachment.core_anchoring import set_core_anchors

from .workflows import MOL

try:
    import nanoCAT
    from nanoCAT.asa.asa import init_asa
    from nanoCAT.mol_bulk import init_lig_bulkiness
    from nanoCAT.bde.bde_workflow import init_bde
    from nanoCAT.ligand_solvation import init_solv
    from nanoCAT.ff.ff_assignment import init_ff_assignment
    from nanoCAT.cdft import init_cdft

    NANO_CAT: Optional[ImportError] = None
except ImportError as ex:
    NANO_CAT = ex

try:
    import dataCAT
    DATA_CAT: Optional[ImportError] = None
except ImportError as ex:
    DATA_CAT = ex

__all__ = ['prep']


def prep(arg: Settings, return_mol: bool = True
         ) -> Optional[Tuple[SettingsDataFrame, SettingsDataFrame, SettingsDataFrame]]:
    """Function that handles all tasks related to the three prep functions.

    * :func:`.prep_core`
    * :func:`.prep_ligand`
    * :func:`.prep_qd`

    Parameters
    ----------
    arg : |plams.Settings|_
        A settings object containing all (optional) arguments.

    return_mol : bool
        If qd_df, core_df & ligand_df should be returned or not.

    Returns
    -------
    |CAT.SettingsDataFrame|_
        Optional: If ``return_mol=True`` return the three QD, core and ligand dataframes.

    """
    # The start
    time_start = time()
    logger.info(f'Starting CAT (version: {__version__})')
    if NANO_CAT is None:
        logger.info(f'The optional Nano-CAT package was successfully found '
                    f'(version: {nanoCAT.__version__})')
    else:
        logger.warning('The optional Nano-CAT package was not found')
        logger.debug(f'{NANO_CAT.__class__.__name__}: {NANO_CAT}', exc_info=True)

    if DATA_CAT is None:
        logger.info(f'The optional Data-CAT package was successfully found '
                    f'(version: {dataCAT.__version__})\n')
    else:
        logger.warning('The optional Data-CAT package was not found')
        logger.debug(f'{DATA_CAT.__class__.__name__}: {DATA_CAT}\n', exc_info=True)

    # Interpret and extract the input settings
    ligand_df, core_df, qd_df = prep_input(arg)

    if qd_df is None:
        # Adds the indices of the core anchor atoms to core.properties.core
        core_df = prep_core(core_df)

        # Optimize the ligands, find functional groups, calculate properties
        # and read/write the results
        ligand_df = prep_ligand(ligand_df)

    # Combine the cores and ligands; analyze the resulting quantum dots
    qd_df = prep_qd(ligand_df, core_df, qd_df)

    # The End
    delta_t = time() - time_start
    logger.info(f'Total elapsed time: {delta_t:.4f} sec')

    if return_mol:
        return qd_df, core_df, ligand_df
    return None


def prep_input(arg: Settings) -> Tuple[SettingsDataFrame, SettingsDataFrame, SettingsDataFrame]:
    """Interpret and extract the input settings. Returns a list of ligands and a list of cores.

    Parameters
    ----------
    |plams.Settings|_
        A settings object containing all (optional) arguments.

    Returns
    -------
    |tuple|_ [|CAT.SettingsDataFrame|_, |CAT.SettingsDataFrame|_, |CAT.SettingsDataFrame|_]
        A tuple containing the ligand, core and qd dataframes.

    """
    # Interpret arguments
    validate_input(arg, validate_only=False)

    # Read the input ligands and cores
    lig_list = read_mol(arg.get('input_ligands'))
    core_list = read_mol(arg.get('input_cores'))
    qd_list = read_mol(arg.get('input_qd'))

    is_qd = True if qd_list is not None else False

    # Raises an error if lig_list or core_list is empty
    if is_qd:
        if not qd_list:
            raise MoleculeError('No valid input quantum dots were found, aborting run')
    else:
        if not lig_list:
            raise MoleculeError('No valid input ligands were found, aborting run')
        elif not core_list:
            raise MoleculeError('No valid input cores were found, aborting run')

    # Store the molecules in dataframes
    columns = pd.MultiIndex.from_tuples([MOL], names=['index', 'sub index'])

    if is_qd:
        ligand_df = core_df = None
        qd_df = SettingsDataFrame(
            index=pd.RangeIndex(len(qd_list)), columns=columns, settings=arg
        )
        qd_df[MOL] = qd_list
    else:
        qd_df = None
        ligand_df = SettingsDataFrame(
            index=pd.RangeIndex(len(lig_list)), columns=columns, settings=arg
        )

        core_df = SettingsDataFrame(
            index=pd.RangeIndex(len(core_list)), columns=columns.copy(), settings=arg
        )

        ligand_df[MOL] = lig_list
        core_df[MOL] = core_list

    return ligand_df, core_df, qd_df


# TODO: Move this function to its own module; this is a workflow and NOT a workflow manager
def prep_core(core_df: SettingsDataFrame) -> SettingsDataFrame:
    """Function that handles the identification and marking of all core anchor atoms.

    Parameters
    ----------
    core_df : |CAT.SettingsDataFrame|_
        A dataframe of core molecules. Molecules are stored in the *mol* column.

    Returns
    -------
    |CAT.SettingsDataFrame|_
        A dataframe of cores with all anchor atoms removed.

    """
    # Unpack arguments
    core_options = core_df.settings.optional.core
    anchor_tup = core_options.anchor[0]
    allignment_tup = core_options.allignment
    subset = core_options.subset

    # Set the core anchors
    idx_tuples = [set_core_anchors(i, anchor_tup, allignment_tup, subset) for i in core_df[MOL]]

    # Create and return a new dataframe
    idx = pd.MultiIndex.from_tuples(idx_tuples, names=['formula', 'anchor'])
    ret = core_df.reindex(idx)
    ret[MOL] = core_df[MOL].values
    return ret


def prep_ligand(ligand_df: SettingsDataFrame) -> SettingsDataFrame:
    """Function that handles all ligand operations.

    * Ligand function group identification
    * Ligand geometry optimization
    * Ligand bulkiness calculations
    * Ligand COSMO-RS calculations
    * Ligand conceptual DFT calculations

    .. _Nano-CAT: https://github.com/nlesc-nano/nano-CAT

    Parameters
    ----------
    ligand_df : |CAT.SettingsDataFrame|_
        A dataframe of ligand molecules. Molecules are stored in the *mol* column.

    Returns
    -------
    |CAT.SettingsDataFrame|_
        A new dataframe containing only valid ligands.

    Raises
    ------
    ImportError
        Raised if a COSMO-RS calculation is attempted without installing the Nano-CAT_ package.

    """
    # Unpack arguments
    forcefield = ligand_df.settings.optional.forcefield
    optimize = ligand_df.settings.optional.ligand.optimize
    crs = ligand_df.settings.optional.ligand.crs
    cdft = ligand_df.settings.optional.ligand.cdft

    # Identify functional groups within the ligand.
    ligand_df = init_ligand_anchoring(ligand_df)

    # Check if any valid functional groups were found
    if not ligand_df[MOL].any():
        raise MoleculeError('No valid functional groups found in any of the ligands, aborting run')

    # Optimize the ligands
    if optimize:
        init_ligand_opt(ligand_df)

        # Remove failed optimizations from the ligand list
        _is_opt = (lig.properties.get('is_opt', False) for lig in ligand_df[MOL])
        is_opt = np.fromiter(_is_opt, count=len(ligand_df), dtype=bool)
        ligand_df = ligand_df.loc[is_opt]
    else:
        for lig in ligand_df[MOL]:
            allign_axis(lig)

    # Perform a COSMO-RS calculation on the ligands
    if crs:
        val_nano_cat("Ligand COSMO-RS calculations require the nano-CAT package")
        init_solv(ligand_df)

    # Assign CHARMM CGenFF atom types to all ligands
    if forcefield:
        val_nano_cat("Automatic ligand forcefield assignment requires MATCH "
                     "(Multipurpose Atom-Typer for CHARMM) and the nano-CAT package")
        init_ff_assignment(ligand_df)

    # Run conceptual DFT calculations
    if cdft:
        val_nano_cat("Ligand conceptual DFT calculations require the nano-CAT package")
        init_cdft(ligand_df)

    return ligand_df


def prep_qd(ligand_df: Optional[SettingsDataFrame],
            core_df: Optional[SettingsDataFrame],
            qd_df: Optional[SettingsDataFrame]) -> SettingsDataFrame:
    """Function that handles all quantum dot (qd, i.e. core + all ligands) operations.

    * Constructing the quantum dots
    * Optimizing the quantum dots
    * Peforming activation strain analyses
    * Dissociating ligands on the quantum dot surface

    .. _Nano-CAT: https://github.com/nlesc-nano/nano-CAT

    Has two accepted signatures:
        * ``ligand_df = core_df = None``: Update an existing quantum dot dataframe (**qd_df**).
        * ``qd_df = None``: Construct a new quantum dot dataframe from
          **ligand_df** and **core_df**.

    Parameters
    ----------
    ligand_df : |CAT.SettingsDataFrame|_
        ``None`` or a dataframe of ligand molecules. Molecules are stored in the *mol* column.

    core_df : |CAT.SettingsDataFrame|_
        ``None`` or a dataframe of core molecules. Molecules are stored in the *mol* column.

    qd_df : |CAT.SettingsDataFrame|_
        ``None`` or a dataframe of quantum dots. Molecules are stored in the *mol* column.

    Returns
    -------
    |CAT.SettingsDataFrame|_
        A dataframe of quantum dots molecules. Molecules are stored in the *mol* column.

    Raises
    ------
    ImportError
        Raised if an activation-strain or ligand dissociation calculation is attempted without
        installing the Nano-CAT_ package.

    """
    # Unpack arguments
    bulk = ligand_df.settings.optional.qd.bulkiness
    optimize = ligand_df.settings.optional.qd.optimize
    forcefield = ligand_df.settings.optional.forcefield
    dissociate = ligand_df.settings.optional.qd.dissociate
    activation_strain = ligand_df.settings.optional.qd.activation_strain
    construct_qd = ligand_df.settings.optional.qd.construct_qd
    multi_ligand = ligand_df.settings.optional.qd.multi_ligand

    # Construct the quantum dot DataFrame
    # If construct_qd is False, construct the dataframe without filling it with quantum dots
    if qd_df is None:  # Construct new quantum dots
        qd_df = init_qd_construction(ligand_df, core_df, construct_qd=construct_qd)
    elif ligand_df is core_df is None:  # Update existing quantum dots
        update_qd_df(qd_df)
    else:
        raise TypeError("Either qd_df must be 'None' or ligand_df "
                        " and core_df must both be 'None'")

    # Start the ligand bulkiness workflow
    if bulk:
        val_nano_cat("Ligand bulkiness calculations require the nano-CAT package")
        init_lig_bulkiness(qd_df, ligand_df, core_df)

    # Skip the actual quantum dot construction
    if not construct_qd:
        return qd_df

    if not qd_df[MOL].any():
        raise MoleculeError('No valid quantum dots found, aborting')

    if forcefield:
        val_nano_cat("Automatic ligand forcefield assignment requires MATCH "
                     "(Multipurpose Atom-Typer for CHARMM) and the nano-CAT package")

    # Optimize the qd with the core frozen
    if optimize:
        init_qd_opt(qd_df)

    if multi_ligand:
        init_multi_ligand(qd_df)

    # Calculate the interaction between ligands on the quantum dot surface
    if activation_strain:
        val_nano_cat("Quantum dot activation-strain calculations require the nano-CAT package")
        init_asa(qd_df)

    # Calculate the interaction between ligands on the quantum dot surface upon removal of CdX2
    if dissociate:
        val_nano_cat("Quantum dot ligand dissociation calculations require the nano-CAT package")
        # Start the BDE calculation
        logger.info('Calculating ligand dissociation energy')
        init_bde(qd_df)

    return qd_df


def val_nano_cat(error_message: Optional[str] = None) -> None:
    """Raise an an :exc:`ImportError` if the module-level constant ``NANO_CAT`` is ``False``."""
    err = error_message if error_message is None else ''
    if NANO_CAT is not None:
        raise ImportError(err) from NANO_CAT
